# Copyright (c) 2018 European Organization for Nuclear Research.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for manipulating NodeGroups via the DB API"""
from oslo_utils import uuidutils

from magnum.common import exception
from magnum.tests.unit.db import base
from magnum.tests.unit.db import utils


class DbNodeGroupTestCase(base.DbTestCase):

    def test_create_nodegroup(self):
        utils.create_test_nodegroup()

    def test_create_nodegroup_already_exists(self):
        utils.create_test_nodegroup()
        self.assertRaises(exception.NodeGroupAlreadyExists,
                          utils.create_test_nodegroup)

    def test_create_nodegroup_same_name_same_cluster(self):
        # NOTE(ttsiouts): Don't allow the same name for nodegroups
        # in the same cluster.
        nodegroup = utils.create_test_nodegroup()
        new = {
            'name': nodegroup.name,
            'id': nodegroup.id + 8,
            'cluster_id': nodegroup.cluster_id
        }
        self.assertRaises(exception.NodeGroupAlreadyExists,
                          utils.create_test_nodegroup, **new)

    def test_create_nodegroup_same_name_different_cluster(self):
        # NOTE(ttsiouts): Verify nodegroups with the same name
        # but in different clusters are allowed.
        nodegroup = utils.create_test_nodegroup()
        new = {
            'name': nodegroup.name,
            'id': nodegroup.id + 8,
            'cluster_id': 'fake-cluster-uuid',
            'uuid': 'fake-nodegroup-uuid',
            'project_id': nodegroup.project_id,
        }
        try:
            utils.create_test_nodegroup(**new)
        except Exception:
            # Something went wrong, just fail the testcase
            self.assertTrue(False)

    def test_get_nodegroup_by_id(self):
        nodegroup = utils.create_test_nodegroup()
        res = self.dbapi.get_nodegroup_by_id(self.context,
                                             nodegroup.cluster_id,
                                             nodegroup.id)
        self.assertEqual(nodegroup.id, res.id)
        self.assertEqual(nodegroup.uuid, res.uuid)

    def test_get_nodegroup_by_name(self):
        nodegroup = utils.create_test_nodegroup()
        res = self.dbapi.get_nodegroup_by_name(self.context,
                                               nodegroup.cluster_id,
                                               nodegroup.name)
        self.assertEqual(nodegroup.name, res.name)
        self.assertEqual(nodegroup.uuid, res.uuid)

    def test_get_cluster_by_uuid(self):
        nodegroup = utils.create_test_nodegroup()
        res = self.dbapi.get_nodegroup_by_uuid(self.context,
                                               nodegroup.cluster_id,
                                               nodegroup.uuid)
        self.assertEqual(nodegroup.id, res.id)
        self.assertEqual(nodegroup.uuid, res.uuid)

    def test_get_nodegroup_that_does_not_exist(self):
        # Create a cluster with no nodegroups
        cluster = utils.create_test_cluster()
        self.assertRaises(exception.NodeGroupNotFound,
                          self.dbapi.get_nodegroup_by_id,
                          self.context, cluster.uuid,  100)
        self.assertRaises(exception.NodeGroupNotFound,
                          self.dbapi.get_nodegroup_by_uuid,
                          self.context, cluster.uuid,
                          '12345678-9999-0000-aaaa-123456789012')
        self.assertRaises(exception.NodeGroupNotFound,
                          self.dbapi.get_nodegroup_by_name,
                          self.context, cluster.uuid, 'not_found')

    def test_get_nodegroups_in_cluster(self):
        uuids_in_cluster = []
        uuids_not_in_cluster = []
        cluster = utils.create_test_cluster(uuid=uuidutils.generate_uuid())
        for i in range(2):
            ng = utils.create_test_nodegroup(uuid=uuidutils.generate_uuid(),
                                             name='test%(id)s' % {'id': i},
                                             cluster_id=cluster.uuid)
            uuids_in_cluster.append(ng.uuid)
        for i in range(2):
            ng = utils.create_test_nodegroup(uuid=uuidutils.generate_uuid(),
                                             name='test%(id)s' % {'id': i},
                                             cluster_id='fake_cluster')
            uuids_not_in_cluster.append(ng.uuid)
        res = self.dbapi.list_cluster_nodegroups(self.context, cluster.uuid)
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids_in_cluster), sorted(res_uuids))
        for uuid in uuids_not_in_cluster:
            self.assertNotIn(uuid, res_uuids)

    def test_get_cluster_list_sorted(self):
        uuids = []
        cluster = utils.create_test_cluster(uuid=uuidutils.generate_uuid())
        for i in range(5):
            ng = utils.create_test_nodegroup(uuid=uuidutils.generate_uuid(),
                                             name='test%(id)s' % {'id': i},
                                             cluster_id=cluster.uuid)
            uuids.append(ng.uuid)
        res = self.dbapi.list_cluster_nodegroups(self.context, cluster.uuid,
                                                 sort_key='uuid')
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids), res_uuids)

        self.assertRaises(exception.InvalidParameterValue,
                          self.dbapi.list_cluster_nodegroups,
                          self.context,
                          cluster.uuid,
                          sort_key='not-there')

    def test_get_nodegroup_list_with_filters(self):
        cluster_dict = utils.get_test_cluster(
            id=1, uuid=uuidutils.generate_uuid())
        cluster = self.dbapi.create_cluster(cluster_dict)

        group1 = utils.create_test_nodegroup(
            name='group-one',
            cluster_id=cluster.uuid,
            flavor_id=1,
            uuid=uuidutils.generate_uuid(),
            node_count=1)
        group2 = utils.create_test_nodegroup(
            name='group-two',
            cluster_id=cluster.uuid,
            flavor_id=1,
            uuid=uuidutils.generate_uuid(),
            node_count=1)
        group3 = utils.create_test_nodegroup(
            name='group-four',
            cluster_id=cluster.uuid,
            flavor_id=2,
            uuid=uuidutils.generate_uuid(),
            node_count=3)

        filters = {'name': 'group-one'}
        res = self.dbapi.list_cluster_nodegroups(
            self.context, cluster.uuid, filters=filters)
        self.assertEqual([group1.id], [r.id for r in res])

        filters = {'node_count': 1}
        res = self.dbapi.list_cluster_nodegroups(
            self.context, cluster.uuid, filters=filters)
        self.assertEqual([group1.id, group2.id], [r.id for r in res])

        filters = {'flavor_id': 2, 'node_count': 3}
        res = self.dbapi.list_cluster_nodegroups(
            self.context, cluster.uuid, filters=filters)
        self.assertEqual([group3.id], [r.id for r in res])

        filters = {'name': 'group-five'}
        res = self.dbapi.list_cluster_nodegroups(
            self.context, cluster.uuid, filters=filters)
        self.assertEqual([], [r.id for r in res])

    def test_destroy_nodegroup(self):
        cluster = utils.create_test_cluster()
        nodegroup = utils.create_test_nodegroup()
        self.assertEqual(nodegroup.uuid, self.dbapi.get_nodegroup_by_uuid(
            self.context, cluster.uuid, nodegroup.uuid).uuid)
        self.dbapi.destroy_nodegroup(cluster.uuid, nodegroup.uuid)

        self.assertRaises(exception.NodeGroupNotFound,
                          self.dbapi.get_nodegroup_by_uuid,
                          self.context, cluster.uuid, nodegroup.uuid)

        self.assertRaises(exception.NodeGroupNotFound,
                          self.dbapi.destroy_nodegroup, cluster.uuid,
                          nodegroup.uuid)

    def test_destroy_nodegroup_by_uuid(self):
        cluster = utils.create_test_cluster()
        nodegroup = utils.create_test_nodegroup()
        self.assertIsNotNone(self.dbapi.get_nodegroup_by_uuid(self.context,
                                                              cluster.uuid,
                                                              nodegroup.uuid))
        self.dbapi.destroy_nodegroup(cluster.uuid, nodegroup.uuid)
        self.assertRaises(exception.NodeGroupNotFound,
                          self.dbapi.get_nodegroup_by_uuid, self.context,
                          cluster.uuid, nodegroup.uuid)

    def test_destroy_cluster_by_uuid_that_does_not_exist(self):
        self.assertRaises(exception.NodeGroupNotFound,
                          self.dbapi.destroy_nodegroup, 'c_uuid',
                          '12345678-9999-0000-aaaa-123456789012')

    def test_update_cluster(self):
        nodegroup = utils.create_test_nodegroup()
        old_flavor = nodegroup.flavor_id
        new_flavor = 5
        self.assertNotEqual(old_flavor, new_flavor)
        res = self.dbapi.update_nodegroup(nodegroup.cluster_id, nodegroup.id,
                                          {'flavor_id': new_flavor})
        self.assertEqual(new_flavor, res.flavor_id)

    def test_update_nodegroup_not_found(self):
        uuid = uuidutils.generate_uuid()
        self.assertRaises(exception.NodeGroupNotFound,
                          self.dbapi.update_nodegroup, "c_uuid", uuid,
                          {'node_count': 5})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pecan
import wsme
from wsme import types as wtypes

from magnum.api.controllers import base
from magnum.api.controllers.v1 import collection
from magnum.api.controllers.v1 import types
from magnum.api import expose
from magnum.api import servicegroup as svcgrp_api
from magnum.common import policy
from magnum import objects
from magnum.objects import fields


class MagnumService(base.APIBase):

    host = wtypes.StringType(min_length=1, max_length=255)
    """Name of the host """

    binary = wtypes.Enum(wtypes.text, *fields.MagnumServiceBinary.ALL)
    """Name of the binary"""

    state = wtypes.Enum(wtypes.text, *fields.MagnumServiceState.ALL)
    """State of the binary"""

    id = wsme.wsattr(wtypes.IntegerType(minimum=1))
    """The id for the healthcheck record """

    report_count = wsme.wsattr(wtypes.IntegerType(minimum=0))
    """The number of times the heartbeat was reported """

    disabled = wsme.wsattr(types.boolean, default=False)
    """If the service is 'disabled' administratively """

    disabled_reason = wtypes.StringType(min_length=0, max_length=255)
    """Reason for disabling """

    def __init__(self, state, **kwargs):
        super(MagnumService, self).__init__()

        self.fields = ['state']
        setattr(self, 'state', state)
        for field in objects.MagnumService.fields:
            self.fields.append(field)
            setattr(self, field, kwargs.get(field, wtypes.Unset))


class MagnumServiceCollection(collection.Collection):

    mservices = [MagnumService]
    """A list containing service objects"""

    def __init__(self, **kwargs):
        super(MagnumServiceCollection, self).__init__()
        self._type = 'mservices'

    @staticmethod
    def convert_db_rec_list_to_collection(servicegroup_api,
                                          rpc_msvcs, **kwargs):
        collection = MagnumServiceCollection()
        collection.mservices = []
        for p in rpc_msvcs:
            alive = servicegroup_api.service_is_up(p)
            state = 'up' if alive else 'down'
            msvc = MagnumService(state, **p.as_dict())
            collection.mservices.append(msvc)
        collection.next = collection.get_next(limit=None, url=None, **kwargs)
        return collection


class MagnumServiceController(base.Controller):
    """REST controller for magnum-services."""

    def __init__(self, **kwargs):
        super(MagnumServiceController, self).__init__()
        self.servicegroup_api = svcgrp_api.ServiceGroup()

    @expose.expose(MagnumServiceCollection)
    @policy.enforce_wsgi("magnum-service")
    def get_all(self):
        """Retrieve a list of magnum-services.

        """
        msvcs = objects.MagnumService.list(pecan.request.context,
                                           limit=None,
                                           marker=None,
                                           sort_key='id',
                                           sort_dir='asc')

        return MagnumServiceCollection.convert_db_rec_list_to_collection(
            self.servicegroup_api, msvcs)



# Copyright (c) 2015 Rackspace Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from magnum.drivers.common import template_def


class ExampleTemplate(template_def.BaseTemplateDefinition):
    provides = [
        {'server_type': 'vm', 'os': 'example', 'coe': 'example_coe'},
        {'server_type': 'vm', 'os': 'example2', 'coe': 'example_coe'},
    ]

    def __init__(self):
        super(ExampleTemplate, self).__init__()

        self.add_output('server_address',
                        bay_attr='api_address')
        self.add_output('node_addresses',
                        bay_attr='node_addresses')

    def template_path(self):
        return os.path.join(os.path.dirname(__file__), 'example.yaml')



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_db.sqlalchemy.migration_cli import manager

import magnum.conf

CONF = magnum.conf.CONF
_MANAGER = None


def get_manager():
    global _MANAGER
    if not _MANAGER:
        alembic_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), 'alembic.ini'))
        migrate_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), 'alembic'))
        migration_config = {'alembic_ini_path': alembic_path,
                            'alembic_repo_path': migrate_path,
                            'db_url': CONF.database.connection}
        _MANAGER = manager.MigrationManager(migration_config)

    return _MANAGER


def version():
    """Current database version.

    :returns: Database version
    :rtype: string
    """
    return get_manager().version()


def upgrade(version):
    """Used for upgrading database.

    :param version: Desired database version
    :type version: string
    """
    version = version or 'head'

    get_manager().upgrade(version)


def stamp(revision):
    """Stamps database with provided revision.

    Don't run any migrations.

    :param revision: Should match one from repository or head - to stamp
                     database with most recent revision
    :type revision: string
    """
    get_manager().stamp(revision)


def revision(message=None, autogenerate=False):
    """Creates template for migration.

    :param message: Text that will be used for migration title
    :type message: string
    :param autogenerate: If True - generates diff based on current database
                         state
    :type autogenerate: bool
    """
    return get_manager().revision(message=message, autogenerate=autogenerate)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Utils for testing the API service.
"""
import datetime

import pytz

from magnum.api.controllers.v1 import bay as bay_controller
from magnum.api.controllers.v1 import baymodel as baymodel_controller
from magnum.api.controllers.v1 import cluster as cluster_controller
from magnum.api.controllers.v1 import cluster_template as cluster_tmp_ctrl
from magnum.api.controllers.v1 import federation as federation_controller
from magnum.tests.unit.db import utils


def remove_internal(values, internal):
    # NOTE(yuriyz): internal attributes should not be posted, except uuid
    int_attr = [attr.lstrip('/') for attr in internal if attr != '/uuid']
    return {k: v for (k, v) in values.items() if k not in int_attr}


def baymodel_post_data(**kw):
    baymodel = utils.get_test_cluster_template(**kw)
    internal = baymodel_controller.BayModelPatchType.internal_attrs()
    return remove_internal(baymodel, internal)


def cluster_template_post_data(**kw):
    cluster_template = utils.get_test_cluster_template(**kw)
    internal = cluster_tmp_ctrl.ClusterTemplatePatchType.internal_attrs()
    return remove_internal(cluster_template, internal)


def bay_post_data(**kw):
    kw.update({'for_api_use': True})
    bay = utils.get_test_cluster(**kw)
    bay['baymodel_id'] = kw.get('baymodel_id', bay['cluster_template_id'])
    bay['bay_create_timeout'] = kw.get('bay_create_timeout', 15)
    del bay['cluster_template_id']
    del bay['create_timeout']
    internal = bay_controller.BayPatchType.internal_attrs()
    return remove_internal(bay, internal)


def cluster_post_data(**kw):
    kw.update({'for_api_use': True})
    cluster = utils.get_test_cluster(**kw)
    cluster['create_timeout'] = kw.get('create_timeout', 15)
    cluster['merge_labels'] = kw.get('merge_labels', False)
    internal = cluster_controller.ClusterPatchType.internal_attrs()
    return remove_internal(cluster, internal)


def cert_post_data(**kw):
    return {
        'cluster_uuid': kw.get('cluster_uuid',
                               '5d12f6fd-a196-4bf0-ae4c-1f639a523a52'),
        'csr': kw.get('csr', 'fake-csr'),
        'pem': kw.get('pem', 'fake-pem')
    }


def quota_post_data(**kw):
    return utils.get_test_quota(**kw)


def mservice_get_data(**kw):
    """Simulate what the RPC layer will get from DB """
    faketime = datetime.datetime(2001, 1, 1, tzinfo=pytz.UTC)
    return {
        'binary': kw.get('binary', 'magnum-conductor'),
        'host': kw.get('host', 'fake-host'),
        'id': kw.get('id', 13),
        'report_count': kw.get('report_count', 13),
        'disabled': kw.get('disabled', False),
        'disabled_reason': kw.get('disabled_reason', None),
        'forced_down': kw.get('forced_down', False),
        'last_seen_at': kw.get('last_seen_at', faketime),
        'created_at': kw.get('created_at', faketime),
        'updated_at': kw.get('updated_at', faketime),
    }


def federation_post_data(**kw):
    federation = utils.get_test_federation(**kw)
    internal = federation_controller.FederationPatchType.internal_attrs()
    return remove_internal(federation, internal)


def nodegroup_post_data(**kw):
    internal = ['/cluster_id', '/project_id', '/node_addresses', '/is_default',
                '/created_at', '/updated_at', '/status', '/status_reason',
                '/version', '/stack_id']
    nodegroup = utils.get_test_nodegroup(**kw)
    nodegroup['merge_labels'] = kw.get('merge_labels', False)
    return remove_internal(nodegroup, internal)



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for manipulating ClusterTemplate via the DB API"""
from oslo_utils import uuidutils
import six

from magnum.common import exception
from magnum.tests.unit.db import base
from magnum.tests.unit.db import utils


class DbClusterTemplateTestCase(base.DbTestCase):

    def test_create_cluster_template(self):
        utils.create_test_cluster_template()

    def test_get_cluster_template_list(self):
        uuids = []
        for i in range(1, 6):
            ct = utils.create_test_cluster_template(
                id=i, uuid=uuidutils.generate_uuid())
            uuids.append(six.text_type(ct['uuid']))
        res = self.dbapi.get_cluster_template_list(self.context)
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids), sorted(res_uuids))

    def test_get_cluster_template_list_sorted(self):
        uuids = []
        for _ in range(5):
            ct = utils.create_test_cluster_template(
                uuid=uuidutils.generate_uuid())
            uuids.append(six.text_type(ct['uuid']))
        res = self.dbapi.get_cluster_template_list(self.context,
                                                   sort_key='uuid')
        res_uuids = [r.uuid for r in res]
        self.assertEqual(sorted(uuids), res_uuids)

        self.assertRaises(exception.InvalidParameterValue,
                          self.dbapi.get_cluster_template_list,
                          self.context,
                          sort_key='foo')

    def test_get_cluster_template_list_with_filters(self):
        ct1 = utils.create_test_cluster_template(
            id=1,
            name='ct-one',
            uuid=uuidutils.generate_uuid(),
            image_id='image1')
        ct2 = utils.create_test_cluster_template(
            id=2,
            name='ct-two',
            uuid=uuidutils.generate_uuid(),
            image_id='image2')

        res = self.dbapi.get_cluster_template_list(self.context,
                                                   filters={'name': 'ct-one'})
        self.assertEqual([ct1['id']], [r.id for r in res])

        res = self.dbapi.get_cluster_template_list(
            self.context, filters={'name': 'bad-name'})
        self.assertEqual([], [r.id for r in res])

        res = self.dbapi.get_cluster_template_list(
            self.context, filters={'image_id': 'image1'})
        self.assertEqual([ct1['id']], [r.id for r in res])

        res = self.dbapi.get_cluster_template_list(
            self.context, filters={'image_id': 'image2'})
        self.assertEqual([ct2['id']], [r.id for r in res])

    def test_get_cluster_template_by_id(self):
        ct = utils.create_test_cluster_template()
        cluster_template = self.dbapi.get_cluster_template_by_id(
            self.context, ct['id'])
        self.assertEqual(ct['uuid'], cluster_template.uuid)

    def test_get_cluster_template_by_id_public(self):
        ct = utils.create_test_cluster_template(user_id='not_me', public=True)
        cluster_template = self.dbapi.get_cluster_template_by_id(
            self.context, ct['id'])
        self.assertEqual(ct['uuid'], cluster_template.uuid)

    def test_get_cluster_template_by_id_hidden(self):
        ct = utils.create_test_cluster_template(user_id='not_me', hidden=True)
        cluster_template = self.dbapi.get_cluster_template_by_id(
            self.context, ct['id'])
        self.assertEqual(ct['uuid'], cluster_template.uuid)

    def test_get_cluster_template_by_uuid(self):
        ct = utils.create_test_cluster_template()
        cluster_template = self.dbapi.get_cluster_template_by_uuid(
            self.context, ct['uuid'])
        self.assertEqual(ct['id'], cluster_template.id)

    def test_get_cluster_template_by_uuid_public(self):
        ct = utils.create_test_cluster_template(user_id='not_me', public=True)
        cluster_template = self.dbapi.get_cluster_template_by_uuid(
            self.context, ct['uuid'])
        self.assertEqual(ct['id'], cluster_template.id)

    def test_get_cluster_template_by_uuid_hidden(self):
        ct = utils.create_test_cluster_template(user_id='not_me', hidden=True)
        cluster_template = self.dbapi.get_cluster_template_by_uuid(
            self.context, ct['uuid'])
        self.assertEqual(ct['id'], cluster_template.id)

    def test_get_cluster_template_that_does_not_exist(self):
        self.assertRaises(exception.ClusterTemplateNotFound,
                          self.dbapi.get_cluster_template_by_id,
                          self.context, 666)

    def test_get_cluster_template_by_name(self):
        ct = utils.create_test_cluster_template()
        res = self.dbapi.get_cluster_template_by_name(self.context, ct['name'])
        self.assertEqual(ct['id'], res.id)
        self.assertEqual(ct['uuid'], res.uuid)

    def test_get_cluster_template_by_name_public(self):
        ct = utils.create_test_cluster_template(user_id='not_me', public=True)
        res = self.dbapi.get_cluster_template_by_name(self.context, ct['name'])
        self.assertEqual(ct['id'], res.id)
        self.assertEqual(ct['uuid'], res.uuid)

    def test_get_cluster_template_by_name_hidden(self):
        ct = utils.create_test_cluster_template(user_id='not_me', hidden=True)
        res = self.dbapi.get_cluster_template_by_name(self.context, ct['name'])
        self.assertEqual(ct['id'], res.id)
        self.assertEqual(ct['uuid'], res.uuid)

    def test_get_cluster_template_by_name_multiple_cluster_template(self):
        utils.create_test_cluster_template(
            id=1, name='ct',
            uuid=uuidutils.generate_uuid(),
            image_id='image1')
        utils.create_test_cluster_template(
            id=2, name='ct',
            uuid=uuidutils.generate_uuid(),
            image_id='image2')
        self.assertRaises(exception.Conflict,
                          self.dbapi.get_cluster_template_by_name,
                          self.context, 'ct')

    def test_get_cluster_template_by_name_not_found(self):
        self.assertRaises(exception.ClusterTemplateNotFound,
                          self.dbapi.get_cluster_template_by_name,
                          self.context, 'not_found')

    def test_get_cluster_template_by_uuid_that_does_not_exist(self):
        self.assertRaises(exception.ClusterTemplateNotFound,
                          self.dbapi.get_cluster_template_by_uuid,
                          self.context,
                          '12345678-9999-0000-aaaa-123456789012')

    def test_update_cluster_template(self):
        ct = utils.create_test_cluster_template()
        res = self.dbapi.update_cluster_template(ct['id'],
                                                 {'name': 'updated-model'})
        self.assertEqual('updated-model', res.name)

    def test_update_cluster_template_that_does_not_exist(self):
        self.assertRaises(exception.ClusterTemplateNotFound,
                          self.dbapi.update_cluster_template, 666,
                          {'name': ''})

    def test_update_cluster_template_uuid(self):
        ct = utils.create_test_cluster_template()
        self.assertRaises(exception.InvalidParameterValue,
                          self.dbapi.update_cluster_template, ct['id'],
                          {'uuid': 'hello'})

    def test_destroy_cluster_template(self):
        ct = utils.create_test_cluster_template()
        self.dbapi.destroy_cluster_template(ct['id'])
        self.assertRaises(exception.ClusterTemplateNotFound,
                          self.dbapi.get_cluster_template_by_id,
                          self.context, ct['id'])

    def test_destroy_cluster_template_by_uuid(self):
        uuid = uuidutils.generate_uuid()
        utils.create_test_cluster_template(uuid=uuid)
        self.assertIsNotNone(self.dbapi.get_cluster_template_by_uuid(
            self.context, uuid))
        self.dbapi.destroy_cluster_template(uuid)
        self.assertRaises(exception.ClusterTemplateNotFound,
                          self.dbapi.get_cluster_template_by_uuid,
                          self.context, uuid)

    def test_destroy_cluster_template_that_does_not_exist(self):
        self.assertRaises(exception.ClusterTemplateNotFound,
                          self.dbapi.destroy_cluster_template, 666)

    def test_destroy_cluster_template_that_referenced_by_clusters(self):
        ct = utils.create_test_cluster_template()
        cluster = utils.create_test_cluster(cluster_template_id=ct['uuid'])
        self.assertEqual(ct['uuid'], cluster.cluster_template_id)
        self.assertRaises(exception.ClusterTemplateReferenced,
                          self.dbapi.destroy_cluster_template, ct['id'])

    def test_create_cluster_template_already_exists(self):
        uuid = uuidutils.generate_uuid()
        utils.create_test_cluster_template(id=1, uuid=uuid)
        self.assertRaises(exception.ClusterTemplateAlreadyExists,
                          utils.create_test_cluster_template,
                          id=2, uuid=uuid)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Add floating_ip_enabled column to baymodel table

Revision ID: b1f612248cab
Revises: 859fb45df249
Create Date: 2016-08-05 15:31:46.203266

"""

# revision identifiers, used by Alembic.
revision = 'b1f612248cab'
down_revision = '859fb45df249'

from alembic import op  # noqa: E402
import sqlalchemy as sa  # noqa: E402


def upgrade():
    op.add_column('baymodel',
                  sa.Column('floating_ip_enabled',
                            sa.Boolean(create_constraint=False),
                            default=True))



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""rename_baymodel_to_clustertemplate

Revision ID: fb03fdef8919
Revises: fcb4efee8f8b
Create Date: 2016-08-31 12:40:31.165817

"""

# revision identifiers, used by Alembic.
revision = 'fb03fdef8919'
down_revision = 'fcb4efee8f8b'

from alembic import op  # noqa: E402


def upgrade():
    op.rename_table('baymodel', 'cluster_template')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import strutils
from oslo_utils import uuidutils
from oslo_versionedobjects import fields

from magnum.db import api as dbapi
from magnum.objects import base
from magnum.objects import fields as m_fields


@base.MagnumObjectRegistry.register
class ClusterTemplate(base.MagnumPersistentObject, base.MagnumObject,
                      base.MagnumObjectDictCompat):
    # Version 1.0: Initial version
    # Version 1.1: Add 'registry_enabled' field
    # Version 1.2: Added 'network_driver' field
    # Version 1.3: Added 'labels' attribute
    # Version 1.4: Added 'insecure' attribute
    # Version 1.5: Changed type of 'coe' from StringField to BayTypeField
    # Version 1.6: Change 'insecure' to 'tls_disabled'
    # Version 1.7: Added 'public' field
    # Version 1.8: Added 'server_type' field
    # Version 1.9: Added 'volume_driver' field
    # Version 1.10: Removed 'ssh_authorized_key' field
    # Version 1.11: Added 'insecure_registry' field
    # Version 1.12: Added 'docker_storage_driver' field
    # Version 1.13: Added 'master_lb_enabled' field
    # Version 1.14: Added 'fixed_subnet' field
    # Version 1.15: Added 'floating_ip_enabled' field
    # Version 1.16: Renamed the class from "BayModel' to 'ClusterTemplate'
    # Version 1.17: 'coe' field type change to ClusterTypeField
    # Version 1.18: DockerStorageDriver is a StringField (was an Enum)
    # Version 1.19: Added 'hidden' field
    # Version 1.20: Added 'tags' field
    VERSION = '1.20'

    dbapi = dbapi.get_instance()

    fields = {
        'id': fields.IntegerField(),
        'uuid': fields.StringField(nullable=True),
        'project_id': fields.StringField(nullable=True),
        'user_id': fields.StringField(nullable=True),
        'name': fields.StringField(nullable=True),
        'image_id': fields.StringField(nullable=True),
        'flavor_id': fields.StringField(nullable=True),
        'master_flavor_id': fields.StringField(nullable=True),
        'keypair_id': fields.StringField(nullable=True),
        'dns_nameserver': fields.StringField(nullable=True),
        'external_network_id': fields.StringField(nullable=True),
        'fixed_network': fields.StringField(nullable=True),
        'fixed_subnet': fields.StringField(nullable=True),
        'network_driver': fields.StringField(nullable=True),
        'volume_driver': fields.StringField(nullable=True),
        'apiserver_port': fields.IntegerField(nullable=True),
        'docker_volume_size': fields.IntegerField(nullable=True),
        'docker_storage_driver': fields.StringField(nullable=True),
        'cluster_distro': fields.StringField(nullable=True),
        'coe': m_fields.ClusterTypeField(nullable=True),
        'http_proxy': fields.StringField(nullable=True),
        'https_proxy': fields.StringField(nullable=True),
        'no_proxy': fields.StringField(nullable=True),
        'registry_enabled': fields.BooleanField(default=False),
        'labels': fields.DictOfStringsField(nullable=True),
        'tls_disabled': fields.BooleanField(default=False),
        'public': fields.BooleanField(default=False),
        'server_type': fields.StringField(nullable=True),
        'insecure_registry': fields.StringField(nullable=True),
        'master_lb_enabled': fields.BooleanField(default=False),
        'floating_ip_enabled': fields.BooleanField(default=True),
        'hidden': fields.BooleanField(default=False),
        'tags': fields.StringField(nullable=True),
    }

    @staticmethod
    def _from_db_object(cluster_template, db_cluster_template):
        """Converts a database entity to a formal object."""
        for field in cluster_template.fields:
            cluster_template[field] = db_cluster_template[field]

        cluster_template.obj_reset_changes()
        return cluster_template

    @staticmethod
    def _from_db_object_list(db_objects, cls, context):
        """Converts a list of database entities to a list of formal objects."""
        return [ClusterTemplate._from_db_object(cls(context), obj) for obj in
                db_objects]

    @base.remotable_classmethod
    def get(cls, context, cluster_template_id):
        """Find and return ClusterTemplate object based on its id or uuid.

        :param cluster_template_id: the id *or* uuid of a ClusterTemplate.
        :param context: Security context
        :returns: a :class:`ClusterTemplate` object.
        """
        if strutils.is_int_like(cluster_template_id):
            return cls.get_by_id(context, cluster_template_id)
        elif uuidutils.is_uuid_like(cluster_template_id):
            return cls.get_by_uuid(context, cluster_template_id)
        else:
            return cls.get_by_name(context, cluster_template_id)

    @base.remotable_classmethod
    def get_by_id(cls, context, cluster_template_id):
        """Find and return ClusterTemplate object based on its integer id.

        :param cluster_template_id: the id of a ClusterTemplate.
        :param context: Security context
        :returns: a :class:`ClusterTemplate` object.
        """
        db_cluster_template = cls.dbapi.get_cluster_template_by_id(
            context, cluster_template_id)
        cluster_template = ClusterTemplate._from_db_object(cls(context),
                                                           db_cluster_template)
        return cluster_template

    @base.remotable_classmethod
    def get_by_uuid(cls, context, uuid):
        """Find and return ClusterTemplate object based on uuid.

        :param uuid: the uuid of a ClusterTemplate.
        :param context: Security context
        :returns: a :class:`ClusterTemplate` object.
        """
        db_cluster_template = cls.dbapi.get_cluster_template_by_uuid(
            context, uuid)
        cluster_template = ClusterTemplate._from_db_object(cls(context),
                                                           db_cluster_template)
        return cluster_template

    @base.remotable_classmethod
    def get_by_name(cls, context, name):
        """Find and return ClusterTemplate object based on name.

        :param name: the name of a ClusterTemplate.
        :param context: Security context
        :returns: a :class:`ClusterTemplate` object.
        """
        db_cluster_template = cls.dbapi.get_cluster_template_by_name(context,
                                                                     name)
        cluster_template = ClusterTemplate._from_db_object(cls(context),
                                                           db_cluster_template)
        return cluster_template

    @base.remotable_classmethod
    def list(cls, context, limit=None, marker=None,
             sort_key=None, sort_dir=None):
        """Return a list of ClusterTemplate objects.

        :param context: Security context.
        :param limit: maximum number of resources to return in a single result.
        :param marker: pagination marker for large data sets.
        :param sort_key: column to sort results by.
        :param sort_dir: direction to sort. "asc" or "desc".
        :returns: a list of :class:`ClusterTemplate` object.

        """
        db_cluster_templates = cls.dbapi.get_cluster_template_list(
            context, limit=limit, marker=marker, sort_key=sort_key,
            sort_dir=sort_dir)
        return ClusterTemplate._from_db_object_list(db_cluster_templates,
                                                    cls, context)

    @base.remotable
    def create(self, context=None):
        """Create a ClusterTemplate record in the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: ClusterTemplate(context)

        """
        values = self.obj_get_changes()
        db_cluster_template = self.dbapi.create_cluster_template(values)
        self._from_db_object(self, db_cluster_template)

    @base.remotable
    def destroy(self, context=None):
        """Delete the ClusterTemplate from the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: ClusterTemplate(context)
        """
        self.dbapi.destroy_cluster_template(self.uuid)
        self.obj_reset_changes()

    @base.remotable
    def save(self, context=None):
        """Save updates to this ClusterTemplate.

        Updates will be made column by column based on the result
        of self.what_changed().

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: ClusterTemplate(context)
        """
        updates = self.obj_get_changes()
        self.dbapi.update_cluster_template(self.uuid, updates)

        self.obj_reset_changes()

    @base.remotable
    def refresh(self, context=None):
        """Loads updates for this ClusterTemplate.

        Loads a ClusterTemplate with the same uuid from the database and
        checks for updated attributes. Updates are applied from
        the loaded ClusterTemplate column by column, if there are any updates.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: ClusterTemplate(context)
        """
        current = self.__class__.get_by_uuid(self._context, uuid=self.uuid)
        for field in self.fields:
            if self.obj_attr_is_set(field) and self[field] != current[field]:
                self[field] = current[field]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""added_tags_to_cluster_template

Revision ID: f1d8b0ab8b8d
Revises: 95096e2334ee
Create Date: 2020-08-26 08:38:11.567618

"""

# revision identifiers, used by Alembic.
revision = 'f1d8b0ab8b8d'
down_revision = '95096e2334ee'

from alembic import op  # noqa: E402
import sqlalchemy as sa  # noqa: E402


def upgrade():
    op.add_column('cluster_template',
                  sa.Column('tags',
                            sa.String(length=255), nullable=True))



# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg


periodic_opts = [
    cfg.BoolOpt('periodic_enable',
                default=True,
                help='Enable periodic tasks.'),
    cfg.IntOpt('periodic_interval_max',
               default=60,
               help='Max interval size between periodic tasks execution in '
                    'seconds.'),
]


def register_opts(conf):
    conf.register_opts(periodic_opts)


def list_opts():
    return {
        "DEFAULT": periodic_opts
    }



# Copyright 2011 Justin Santa Barbara
# Copyright 2012 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import os
import os.path
import shutil
import tempfile
from unittest import mock

from oslo_concurrency import processutils
from oslo_utils import netutils

from magnum.common import exception
from magnum.common import utils
import magnum.conf
from magnum.tests import base

CONF = magnum.conf.CONF


class UtilsTestCase(base.TestCase):

    def test_get_k8s_quantity(self):
        self.assertEqual(1024000.0, utils.get_k8s_quantity('1000Ki'))
        self.assertEqual(0.001, utils.get_k8s_quantity('1E-3'))
        self.assertEqual(0.5, utils.get_k8s_quantity('0.0005k'))
        self.assertEqual(0.5, utils.get_k8s_quantity('500m'))
        self.assertEqual(1300000.0, utils.get_k8s_quantity('1.3E+6'))
        self.assertEqual(1300000.0, utils.get_k8s_quantity('1.3E6'))
        self.assertRaises(exception.UnsupportedK8sQuantityFormat,
                          utils.get_k8s_quantity, '1E1E')

    def test_get_docker_quantity(self):
        self.assertEqual(512, utils.get_docker_quantity('512'))
        self.assertEqual(512, utils.get_docker_quantity('512b'))
        self.assertEqual(512 * 1024, utils.get_docker_quantity('512k'))
        self.assertEqual(512 * 1024 * 1024, utils.get_docker_quantity('512m'))
        self.assertEqual(512 * 1024 * 1024 * 1024,
                         utils.get_docker_quantity('512g'))
        self.assertRaises(exception.UnsupportedDockerQuantityFormat,
                          utils.get_docker_quantity, '512bb')
        self.assertRaises(exception.UnsupportedDockerQuantityFormat,
                          utils.get_docker_quantity, '512B')

    def test_get_openstasck_ca(self):
        # openstack_ca_file is empty
        self.assertEqual('', utils.get_openstack_ca())

        # openstack_ca_file is set but the file doesn't exist
        CONF.set_override('openstack_ca_file',
                          '/tmp/invalid-ca.pem',
                          group='drivers')
        self.assertRaises(IOError, utils.get_openstack_ca)

        # openstack_ca_file is set and the file exists
        CONF.set_override('openstack_ca_file',
                          '/tmp/invalid-ca.pem',
                          group='drivers')
        with mock.patch('magnum.common.utils.open',
                        mock.mock_open(read_data="CERT"), create=True):
            self.assertEqual('CERT', utils.get_openstack_ca())


class ExecuteTestCase(base.TestCase):

    def test_retry_on_failure(self):
        fd, tmpfilename = tempfile.mkstemp()
        _, tmpfilename2 = tempfile.mkstemp()
        try:
            fp = os.fdopen(fd, 'w+')
            fp.write('''#!/bin/sh
# If stdin fails to get passed during one of the runs, make a note.
if ! grep -q foo
then
    echo 'failure' > "$1"
fi
# If stdin has failed to get passed during this or a previous run, exit early.
if grep failure "$1"
then
    exit 1
fi
runs="$(cat $1)"
if [ -z "$runs" ]
then
    runs=0
fi
runs=$(($runs + 1))
echo $runs > "$1"
exit 1
''')
            fp.close()
            os.chmod(tmpfilename, 0o755)
            try:
                self.assertRaises(processutils.ProcessExecutionError,
                                  utils.execute,
                                  tmpfilename, tmpfilename2, attempts=10,
                                  process_input=b'foo',
                                  delay_on_retry=False)
            except OSError as e:
                if e.errno == errno.EACCES:
                    self.skipTest("Permissions error detected. "
                                  "Are you running with a noexec /tmp?")
                else:
                    raise
            with open(tmpfilename2, 'r') as fp:
                runs = fp.read()
            self.assertNotEqual(runs.strip(), 'failure', 'stdin did not '
                                'always get passed '
                                'correctly')
            runs = int(runs.strip())
            self.assertEqual(10, runs,
                             'Ran %d times instead of 10.' % runs)
        finally:
            os.unlink(tmpfilename)
            os.unlink(tmpfilename2)

    def test_unknown_kwargs_raises_error(self):
        self.assertRaises(processutils.UnknownArgumentError,
                          utils.execute,
                          '/usr/bin/env', 'true',
                          this_is_not_a_valid_kwarg=True)

    def test_check_exit_code_boolean(self):
        utils.execute('/usr/bin/env', 'false', check_exit_code=False)
        self.assertRaises(processutils.ProcessExecutionError,
                          utils.execute,
                          '/usr/bin/env', 'false', check_exit_code=True)

    def test_no_retry_on_success(self):
        fd, tmpfilename = tempfile.mkstemp()
        _, tmpfilename2 = tempfile.mkstemp()
        try:
            fp = os.fdopen(fd, 'w+')
            fp.write('''#!/bin/sh
# If we've already run, bail out.
grep -q foo "$1" && exit 1
# Mark that we've run before.
echo foo > "$1"
# Check that stdin gets passed correctly.
grep foo
''')
            fp.close()
            os.chmod(tmpfilename, 0o755)
            try:
                utils.execute(tmpfilename,
                              tmpfilename2,
                              process_input=b'foo',
                              attempts=2)
            except OSError as e:
                if e.errno == errno.EACCES:
                    self.skipTest("Permissions error detected. "
                                  "Are you running with a noexec /tmp?")
                else:
                    raise
        finally:
            os.unlink(tmpfilename)
            os.unlink(tmpfilename2)

    @mock.patch.object(processutils, 'execute')
    @mock.patch.object(os.environ, 'copy', return_value={})
    def test_execute_use_standard_locale_no_env_variables(self, env_mock,
                                                          execute_mock):
        utils.execute('foo', use_standard_locale=True)
        execute_mock.assert_called_once_with('foo',
                                             env_variables={'LC_ALL': 'C'})

    @mock.patch.object(processutils, 'execute')
    def test_execute_use_standard_locale_with_env_variables(self,
                                                            execute_mock):
        utils.execute('foo', use_standard_locale=True,
                      env_variables={'foo': 'bar'})
        execute_mock.assert_called_once_with('foo',
                                             env_variables={'LC_ALL': 'C',
                                                            'foo': 'bar'})

    @mock.patch.object(processutils, 'execute')
    def test_execute_not_use_standard_locale(self, execute_mock):
        utils.execute('foo', use_standard_locale=False,
                      env_variables={'foo': 'bar'})
        execute_mock.assert_called_once_with('foo',
                                             env_variables={'foo': 'bar'})

    def test_execute_get_root_helper(self):
        with mock.patch.object(processutils, 'execute') as execute_mock:
            helper = utils._get_root_helper()
            utils.execute('foo', run_as_root=True)
            execute_mock.assert_called_once_with('foo', run_as_root=True,
                                                 root_helper=helper)

    def test_execute_without_root_helper(self):
        with mock.patch.object(processutils, 'execute') as execute_mock:
            utils.execute('foo', run_as_root=False)
            execute_mock.assert_called_once_with('foo', run_as_root=False)

    def test_validate_and_normalize_mac(self):
        mac = 'AA:BB:CC:DD:EE:FF'
        with mock.patch.object(netutils, 'is_valid_mac') as m_mock:
            m_mock.return_value = True
            self.assertEqual(mac.lower(),
                             utils.validate_and_normalize_mac(mac))

    def test_validate_and_normalize_mac_invalid_format(self):
        with mock.patch.object(netutils, 'is_valid_mac') as m_mock:
            m_mock.return_value = False
            self.assertRaises(exception.InvalidMAC,
                              utils.validate_and_normalize_mac, 'invalid-mac')

    def test_safe_rstrip(self):
        value = '/test/'
        rstripped_value = '/test'
        not_rstripped = '/'

        self.assertEqual(rstripped_value, utils.safe_rstrip(value, '/'))
        self.assertEqual(not_rstripped, utils.safe_rstrip(not_rstripped, '/'))

    def test_safe_rstrip_not_raises_exceptions(self):
        # Supplying an integer should normally raise an exception because it
        # does not save the rstrip() method.
        value = 10

        # In the case of raising an exception safe_rstrip() should return the
        # original value.
        self.assertEqual(value, utils.safe_rstrip(value))


class TempFilesTestCase(base.TestCase):

    def test_tempdir(self):

        dirname = None
        with utils.tempdir() as tempdir:
            self.assertTrue(os.path.isdir(tempdir))
            dirname = tempdir
        self.assertFalse(os.path.exists(dirname))

    @mock.patch.object(shutil, 'rmtree')
    @mock.patch.object(tempfile, 'mkdtemp')
    def test_tempdir_mocked(self, mkdtemp_mock, rmtree_mock):

        self.config(tempdir='abc')
        mkdtemp_mock.return_value = 'temp-dir'
        kwargs = {'a': 'b'}

        with utils.tempdir(**kwargs) as tempdir:
            self.assertEqual('temp-dir', tempdir)
            tempdir_created = tempdir

        mkdtemp_mock.assert_called_once_with(**kwargs)
        rmtree_mock.assert_called_once_with(tempdir_created)

    @mock.patch.object(utils, 'LOG')
    @mock.patch.object(shutil, 'rmtree')
    @mock.patch.object(tempfile, 'mkdtemp')
    def test_tempdir_mocked_error_on_rmtree(self, mkdtemp_mock, rmtree_mock,
                                            log_mock):

        self.config(tempdir='abc')
        mkdtemp_mock.return_value = 'temp-dir'
        rmtree_mock.side_effect = OSError

        with utils.tempdir() as tempdir:
            self.assertEqual('temp-dir', tempdir)
            tempdir_created = tempdir

        rmtree_mock.assert_called_once_with(tempdir_created)
        self.assertTrue(log_mock.error.called)


class GeneratePasswordTestCase(base.TestCase):
    def test_generate_password(self):
        password = utils.generate_password(length=12)
        self.assertTrue([c for c in password if c in '0123456789'])
        self.assertTrue([c for c in password
                         if c in 'abcdefghijklmnopqrstuvwxyz'])
        self.assertTrue([c for c in password
                         if c in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'])



# Copyright (c) 2018 European Organization for Nuclear Research.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import uuidutils
from testtools.matchers import HasLength

from magnum import objects
from magnum.tests.unit.db import base
from magnum.tests.unit.db import utils


class TestNodeGroupObject(base.DbTestCase):

    def setUp(self):
        super(TestNodeGroupObject, self).setUp()
        self.fake_nodegroup = utils.get_test_nodegroup()
        self.fake_nodegroup['docker_volume_size'] = 3
        self.fake_nodegroup['labels'] = {}

    def test_get_by_id(self):
        nodegroup_id = self.fake_nodegroup['id']
        cluster_id = self.fake_nodegroup['cluster_id']
        with mock.patch.object(self.dbapi, 'get_nodegroup_by_id',
                               autospec=True) as mock_get_nodegroup:
            mock_get_nodegroup.return_value = self.fake_nodegroup
            nodegroup = objects.NodeGroup.get(self.context, cluster_id,
                                              nodegroup_id)
            mock_get_nodegroup.assert_called_once_with(self.context,
                                                       cluster_id,
                                                       nodegroup_id)
            self.assertEqual(self.context, nodegroup._context)

    def test_get_by_uuid(self):
        uuid = self.fake_nodegroup['uuid']
        cluster_id = self.fake_nodegroup['cluster_id']
        with mock.patch.object(self.dbapi, 'get_nodegroup_by_uuid',
                               autospec=True) as mock_get_nodegroup:
            mock_get_nodegroup.return_value = self.fake_nodegroup
            nodegroup = objects.NodeGroup.get(self.context, cluster_id, uuid)
            mock_get_nodegroup.assert_called_once_with(self.context,
                                                       cluster_id, uuid)
            self.assertEqual(self.context, nodegroup._context)

    def test_get_by_name(self):
        name = self.fake_nodegroup['name']
        cluster_id = self.fake_nodegroup['cluster_id']
        with mock.patch.object(self.dbapi, 'get_nodegroup_by_name',
                               autospec=True) as mock_get_nodegroup:
            mock_get_nodegroup.return_value = self.fake_nodegroup
            nodegroup = objects.NodeGroup.get(self.context, cluster_id, name)
            mock_get_nodegroup.assert_called_once_with(self.context,
                                                       cluster_id, name)
            self.assertEqual(self.context, nodegroup._context)

    def test_list(self):
        cluster_id = self.fake_nodegroup['cluster_id']
        with mock.patch.object(self.dbapi, 'list_cluster_nodegroups',
                               autospec=True) as mock_get_list:
            mock_get_list.return_value = [self.fake_nodegroup]
            nodegroups = objects.NodeGroup.list(self.context, cluster_id)
            self.assertEqual(1, mock_get_list.call_count)
            mock_get_list.assert_called_once_with(
                self.context, cluster_id, limit=None, marker=None,
                filters=None, sort_dir=None, sort_key=None)
            self.assertThat(nodegroups, HasLength(1))
            self.assertIsInstance(nodegroups[0], objects.NodeGroup)
            self.assertEqual(self.context, nodegroups[0]._context)

    def test_list_with_filters(self):
        cluster_id = self.fake_nodegroup['cluster_id']
        with mock.patch.object(self.dbapi, 'list_cluster_nodegroups',
                               autospec=True) as mock_get_list:
            mock_get_list.return_value = [self.fake_nodegroup]
            filters = {'name': self.fake_nodegroup['name']}
            nodegroups = objects.NodeGroup.list(self.context, cluster_id,
                                                filters=filters)
            self.assertEqual(1, mock_get_list.call_count)
            mock_get_list.assert_called_once_with(
                self.context, cluster_id, limit=None, marker=None,
                filters=filters, sort_dir=None, sort_key=None)
            self.assertThat(nodegroups, HasLength(1))
            self.assertIsInstance(nodegroups[0], objects.NodeGroup)
            self.assertEqual(self.context, nodegroups[0]._context)

    def test_create(self):
        with mock.patch.object(self.dbapi, 'create_nodegroup',
                               autospec=True) as mock_create_nodegroup:
            mock_create_nodegroup.return_value = self.fake_nodegroup
            nodegroup = objects.NodeGroup(self.context, **self.fake_nodegroup)
            nodegroup.create()
            mock_create_nodegroup.assert_called_once_with(self.fake_nodegroup)
            self.assertEqual(self.context, nodegroup._context)

    def test_destroy(self):
        uuid = self.fake_nodegroup['uuid']
        cluster_id = self.fake_nodegroup['cluster_id']
        with mock.patch.object(self.dbapi, 'get_nodegroup_by_uuid',
                               autospec=True) as mock_get_nodegroup:
            mock_get_nodegroup.return_value = self.fake_nodegroup
            with mock.patch.object(self.dbapi, 'destroy_nodegroup',
                                   autospec=True) as mock_destroy_nodegroup:
                nodegroup = objects.NodeGroup.get_by_uuid(self.context,
                                                          cluster_id,
                                                          uuid)
                nodegroup.destroy()
                mock_get_nodegroup.assert_called_once_with(self.context,
                                                           cluster_id,
                                                           uuid)
                mock_destroy_nodegroup.assert_called_once_with(cluster_id,
                                                               uuid)
                self.assertEqual(self.context, nodegroup._context)

    def test_save(self):
        uuid = self.fake_nodegroup['uuid']
        cluster_id = self.fake_nodegroup['cluster_id']
        with mock.patch.object(self.dbapi, 'get_nodegroup_by_uuid',
                               autospec=True) as mock_get_nodegroup:
            mock_get_nodegroup.return_value = self.fake_nodegroup
            with mock.patch.object(self.dbapi, 'update_nodegroup',
                                   autospec=True) as mock_update_nodegroup:
                nodegroup = objects.NodeGroup.get_by_uuid(self.context,
                                                          cluster_id,
                                                          uuid)
                nodegroup.node_count = 10
                nodegroup.save()

                mock_get_nodegroup.assert_called_once_with(self.context,
                                                           cluster_id,
                                                           uuid)
                expected_changes = {
                    'node_count': 10,
                }
                mock_update_nodegroup.assert_called_once_with(
                    cluster_id, uuid, expected_changes)
                self.assertEqual(self.context, nodegroup._context)

    def test_refresh(self):
        uuid = self.fake_nodegroup['uuid']
        cluster_id = self.fake_nodegroup['cluster_id']
        new_uuid = uuidutils.generate_uuid()
        returns = [dict(self.fake_nodegroup, uuid=uuid),
                   dict(self.fake_nodegroup, uuid=new_uuid)]
        expected = [mock.call(self.context, cluster_id, uuid),
                    mock.call(self.context, cluster_id, uuid)]
        with mock.patch.object(self.dbapi, 'get_nodegroup_by_uuid',
                               side_effect=returns,
                               autospec=True) as mock_get_nodegroup:
            nodegroup = objects.NodeGroup.get_by_uuid(self.context,
                                                      cluster_id,
                                                      uuid)
            self.assertEqual(uuid, nodegroup.uuid)
            nodegroup.refresh()
            self.assertEqual(new_uuid, nodegroup.uuid)
            self.assertEqual(expected, mock_get_nodegroup.call_args_list)
            self.assertEqual(self.context, nodegroup._context)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Starter script for magnum-driver-manage."""
import sys

from cliff import app
from cliff import commandmanager
from cliff import lister

import magnum.conf
from magnum.drivers.common import driver
from magnum import version

CONF = magnum.conf.CONF


class DriverList(lister.Lister):
    """List templates"""

    def _print_rows(self, parsed_args, rows):
        fields = ['name']
        field_labels = ['Name']

        if parsed_args.details:
            fields.extend(['server_type', 'os', 'coe'])
            field_labels.extend(['Server_Type', 'OS', 'COE'])
        if parsed_args.paths:
            fields.append('path')
            field_labels.append('Template Path')
        return field_labels, [tuple([row[field] for field in fields])
                              for row in rows]

    def get_parser(self, prog_name):
        parser = super(DriverList, self).get_parser(prog_name)
        parser.add_argument('-d', '--details',
                            action='store_true',
                            dest='details',
                            help=('display the cluster types provided by '
                                  'each template'))
        parser.add_argument('-p', '--paths',
                            action='store_true',
                            dest='paths',
                            help='display the path to each template file')

        return parser

    def take_action(self, parsed_args):
        rows = []

        for entry_point, cls in driver.Driver.load_entry_points():
            name = entry_point.name
            definition = cls().get_template_definition()
            template = dict(name=name, path=definition.template_path)

            if parsed_args.details:
                for cluster_type in cls().provides:
                    row = dict()
                    row.update(template)
                    row.update(cluster_type)
                    rows.append(row)
            else:
                rows.append(template)
        return self._print_rows(parsed_args, rows)


class DriverCommandManager(commandmanager.CommandManager):
    COMMANDS = {
        "list-drivers": DriverList,
    }

    def load_commands(self, namespace):
        for name, command_class in self.COMMANDS.items():
            self.add_command(name, command_class)


class DriverManager(app.App):
    def __init__(self):
        super(DriverManager, self).__init__(
            description='Magnum Driver Manager',
            version=version.version_info,
            command_manager=DriverCommandManager('magnum'),
            deferred_help=True)


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    CONF([],
         project='magnum',
         version=version.version_info.release_string())
    return DriverManager().run(args)



# Copyright 2015 Huawei Technologies Co.,LTD.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import tempfile
from unittest import mock

from requests_mock.contrib import fixture

from magnum.common import exception
from magnum.conductor import scale_manager
from magnum.drivers.common.k8s_scale_manager import K8sScaleManager
from magnum.drivers.mesos_ubuntu_v1.scale_manager import MesosScaleManager
from magnum.tests import base


class TestScaleManager(base.TestCase):

    def _test_get_removal_nodes(
            self, mock_get_hosts, mock_get_num_of_removal,
            mock_is_scale_down, mock_get_by_uuid, is_scale_down,
            num_of_removal, all_hosts, container_hosts,
            expected_removal_hosts):

        mock_is_scale_down.return_value = is_scale_down
        mock_get_num_of_removal.return_value = num_of_removal

        mock_get_hosts.return_value = container_hosts

        mock_heat_output = mock.MagicMock()
        mock_heat_output.get_output_value.return_value = all_hosts

        mock_stack = mock.MagicMock()
        mock_heat_client = mock.MagicMock()
        mock_osc = mock.MagicMock()
        mock_heat_client.stacks.get.return_value = mock_stack
        mock_osc.heat.return_value = mock_heat_client

        mock_context = mock.MagicMock()
        mock_cluster = mock.MagicMock()
        scale_mgr = scale_manager.ScaleManager(mock_context, mock_osc,
                                               mock_cluster)

        if expected_removal_hosts is None:
            self.assertRaises(exception.MagnumException,
                              scale_mgr.get_removal_nodes, mock_heat_output)
        else:
            removal_hosts = scale_mgr.get_removal_nodes(mock_heat_output)
            self.assertEqual(expected_removal_hosts, removal_hosts)
            if num_of_removal > 0:
                mock_get_hosts.assert_called_once_with(mock_context,
                                                       mock_cluster)

    @mock.patch('magnum.objects.Cluster.get_by_uuid')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager._is_scale_down')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager.'
                '_get_num_of_removal')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager.'
                '_get_hosts_with_container')
    def test_get_removal_nodes_no_container_host(
            self, mock_get_hosts, mock_get_num_of_removal,
            mock_is_scale_down, mock_get_by_uuid):

        is_scale_down = True
        num_of_removal = 1
        all_hosts = ['10.0.0.3']
        container_hosts = set()
        expected_removal_hosts = ['10.0.0.3']
        self._test_get_removal_nodes(
            mock_get_hosts, mock_get_num_of_removal, mock_is_scale_down,
            mock_get_by_uuid, is_scale_down, num_of_removal, all_hosts,
            container_hosts, expected_removal_hosts)

    @mock.patch('magnum.objects.Cluster.get_by_uuid')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager._is_scale_down')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager.'
                '_get_num_of_removal')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager.'
                '_get_hosts_with_container')
    def test_get_removal_nodes_one_container_host(
            self, mock_get_hosts, mock_get_num_of_removal,
            mock_is_scale_down, mock_get_by_uuid):

        is_scale_down = True
        num_of_removal = 1
        all_hosts = ['10.0.0.3', '10.0.0.4']
        container_hosts = set(['10.0.0.3'])
        expected_removal_hosts = ['10.0.0.4']
        self._test_get_removal_nodes(
            mock_get_hosts, mock_get_num_of_removal, mock_is_scale_down,
            mock_get_by_uuid, is_scale_down, num_of_removal, all_hosts,
            container_hosts, expected_removal_hosts)

    @mock.patch('magnum.objects.Cluster.get_by_uuid')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager._is_scale_down')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager.'
                '_get_num_of_removal')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager.'
                '_get_hosts_with_container')
    def test_get_removal_nodes_two_container_hosts(
            self, mock_get_hosts, mock_get_num_of_removal,
            mock_is_scale_down, mock_get_by_uuid):

        is_scale_down = True
        num_of_removal = 1
        all_hosts = ['10.0.0.3', '10.0.0.4']
        container_hosts = set(['10.0.0.3', '10.0.0.4'])
        expected_removal_hosts = []
        self._test_get_removal_nodes(
            mock_get_hosts, mock_get_num_of_removal, mock_is_scale_down,
            mock_get_by_uuid, is_scale_down, num_of_removal, all_hosts,
            container_hosts, expected_removal_hosts)

    @mock.patch('magnum.objects.Cluster.get_by_uuid')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager._is_scale_down')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager.'
                '_get_num_of_removal')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager.'
                '_get_hosts_with_container')
    def test_get_removal_nodes_three_container_hosts(
            self, mock_get_hosts, mock_get_num_of_removal,
            mock_is_scale_down, mock_get_by_uuid):

        is_scale_down = True
        num_of_removal = 1
        all_hosts = ['10.0.0.3', '10.0.0.4']
        container_hosts = set(['10.0.0.3', '10.0.0.4', '10.0.0.5'])
        expected_removal_hosts = []
        self._test_get_removal_nodes(
            mock_get_hosts, mock_get_num_of_removal, mock_is_scale_down,
            mock_get_by_uuid, is_scale_down, num_of_removal, all_hosts,
            container_hosts, expected_removal_hosts)

    @mock.patch('magnum.objects.Cluster.get_by_uuid')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager._is_scale_down')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager.'
                '_get_num_of_removal')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager.'
                '_get_hosts_with_container')
    def test_get_removal_nodes_scale_up(
            self, mock_get_hosts, mock_get_num_of_removal,
            mock_is_scale_down, mock_get_by_uuid):

        is_scale_down = False
        num_of_removal = -1
        all_hosts = ['10.0.0.3', '10.0.0.4']
        container_hosts = set()
        expected_removal_hosts = []
        self._test_get_removal_nodes(
            mock_get_hosts, mock_get_num_of_removal, mock_is_scale_down,
            mock_get_by_uuid, is_scale_down, num_of_removal, all_hosts,
            container_hosts, expected_removal_hosts)

    @mock.patch('magnum.objects.Cluster.get_by_uuid')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager._is_scale_down')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager.'
                '_get_num_of_removal')
    @mock.patch('magnum.conductor.scale_manager.ScaleManager.'
                '_get_hosts_with_container')
    def test_get_removal_nodes_with_none_hosts(
            self, mock_get_hosts, mock_get_num_of_removal,
            mock_is_scale_down, mock_get_by_uuid):

        is_scale_down = True
        num_of_removal = 1
        all_hosts = None
        container_hosts = set()
        expected_removal_hosts = None
        self._test_get_removal_nodes(
            mock_get_hosts, mock_get_num_of_removal, mock_is_scale_down,
            mock_get_by_uuid, is_scale_down, num_of_removal, all_hosts,
            container_hosts, expected_removal_hosts)


class TestK8sScaleManager(base.TestCase):

    def setUp(self):
        super(TestK8sScaleManager, self).setUp()
        self.requests_mock = self.useFixture(fixture.Fixture())

    @mock.patch('magnum.objects.Cluster.get_by_uuid')
    @mock.patch('magnum.conductor.k8s_api.create_client_files')
    def test_get_hosts_with_container(self, mock_create_client_files, mock_get):
        mock_cluster = mock.MagicMock()
        mock_cluster.api_address = "https://foobar.com:6443"

        mock_create_client_files.return_value = (
            tempfile.NamedTemporaryFile(),
            tempfile.NamedTemporaryFile(),
            tempfile.NamedTemporaryFile()
        )

        self.requests_mock.register_uri(
            'GET',
            f"{mock_cluster.api_address}/api/v1/namespaces/default/pods",
            json={
                'items': [
                    {
                        'spec': {
                            'node_name': 'node1',
                        }
                    },
                    {
                        'spec': {
                            'node_name': 'node2',
                        }
                    }
                ]
            },
        )

        mgr = K8sScaleManager(
            mock.MagicMock(), mock.MagicMock(), mock.MagicMock())
        hosts = mgr._get_hosts_with_container(
            mock.MagicMock(), mock_cluster)
        self.assertEqual(hosts, {'node1', 'node2'})


class TestMesosScaleManager(base.TestCase):

    @mock.patch('magnum.objects.Cluster.get_by_uuid')
    @mock.patch('marathon.MarathonClient')
    @mock.patch('marathon.MarathonClient.list_tasks')
    def test_get_hosts_with_container(self, mock_list_tasks,
                                      mock_client,  mock_get):
        task_1 = mock.MagicMock()
        task_1.host = 'node1'
        task_2 = mock.MagicMock()
        task_2.host = 'node2'
        tasks = [task_1, task_2]
        mock_list_tasks.return_value = tasks

        mgr = MesosScaleManager(
            mock.MagicMock(), mock.MagicMock(), mock.MagicMock())
        hosts = mgr._get_hosts_with_container(
            mock.MagicMock(), mock.MagicMock())
        self.assertEqual(hosts, {'node1', 'node2'})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from unittest import mock

from oslo_config import cfg
from oslo_utils import uuidutils

from magnum.api.controllers.v1 import federation as api_federation
from magnum.conductor import api as rpcapi
import magnum.conf
from magnum import objects
from magnum.tests import base
from magnum.tests.unit.api import base as api_base
from magnum.tests.unit.api import utils as apiutils
from magnum.tests.unit.objects import utils as obj_utils

CONF = magnum.conf.CONF


class TestFederationObject(base.TestCase):
    def test_federation_init(self):
        fed_dict = apiutils.federation_post_data()
        fed_dict['uuid'] = uuidutils.generate_uuid()
        federation = api_federation.Federation(**fed_dict)
        self.assertEqual(fed_dict['uuid'], federation.uuid)


class TestListFederation(api_base.FunctionalTest):
    def setUp(self):
        super(TestListFederation, self).setUp()

    def test_empty(self):
        response = self.get_json('/federations')
        self.assertEqual(response['federations'], [])

    def test_one(self):
        federation = obj_utils.create_test_federation(
            self.context, uuid=uuidutils.generate_uuid())
        response = self.get_json('/federations')
        self.assertEqual(federation.uuid, response['federations'][0]['uuid'])

    def test_get_one(self):
        federation = obj_utils.create_test_federation(
            self.context, uuid=uuidutils.generate_uuid())
        response = self.get_json('/federations/%s' % federation['uuid'])
        self.assertTrue(response['uuid'], federation.uuid)

    def test_get_one_by_name(self):
        federation = obj_utils.create_test_federation(
            self.context, uuid=uuidutils.generate_uuid())
        response = self.get_json('/federations/%s' % federation['name'])
        self.assertTrue(response['uuid'], federation.uuid)

    def test_get_one_by_name_not_found(self):
        response = self.get_json('/federations/not_found', expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_get_one_by_uuid(self):
        temp_uuid = uuidutils.generate_uuid()
        federation = obj_utils.create_test_federation(self.context,
                                                      uuid=temp_uuid)
        response = self.get_json('/federations/%s' % temp_uuid)
        self.assertTrue(response['uuid'], federation.uuid)

    def test_get_one_by_uuid_not_found(self):
        temp_uuid = uuidutils.generate_uuid()
        response = self.get_json('/federations/%s' % temp_uuid,
                                 expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_get_one_by_name_multiple_federation(self):
        obj_utils.create_test_federation(self.context, name='test_federation',
                                         uuid=uuidutils.generate_uuid())
        obj_utils.create_test_federation(self.context, name='test_federation',
                                         uuid=uuidutils.generate_uuid())
        response = self.get_json('/federations/test_federation',
                                 expect_errors=True)
        self.assertEqual(409, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_get_all_with_pagination_marker(self):
        federation_list = []
        for id_ in range(4):
            federation = obj_utils.create_test_federation(
                self.context, id=id_, uuid=uuidutils.generate_uuid())
            federation_list.append(federation)

        response = self.get_json(
            '/federations?limit=3&marker=%s' % federation_list[2].uuid)
        self.assertEqual(1, len(response['federations']))
        self.assertEqual(federation_list[-1].uuid,
                         response['federations'][0]['uuid'])

    def test_detail(self):
        federation = obj_utils.create_test_federation(
            self.context, uuid=uuidutils.generate_uuid())
        response = self.get_json('/federations/detail')
        self.assertEqual(federation.uuid, response['federations'][0]["uuid"])

    def test_detail_with_pagination_marker(self):
        federation_list = []
        for id_ in range(4):
            federation = obj_utils.create_test_federation(
                self.context, id=id_, uuid=uuidutils.generate_uuid())
            federation_list.append(federation)

        response = self.get_json(
            '/federations/detail?limit=3&marker=%s' % federation_list[2].uuid)
        self.assertEqual(1, len(response['federations']))
        self.assertEqual(federation_list[-1].uuid,
                         response['federations'][0]['uuid'])

    def test_detail_against_single(self):
        federation = obj_utils.create_test_federation(
            self.context, uuid=uuidutils.generate_uuid())
        response = self.get_json(
            '/federations/%s/detail' % federation['uuid'], expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_many(self):
        federation_list = []
        for id_ in range(5):
            temp_uuid = uuidutils.generate_uuid()
            federation = obj_utils.create_test_federation(
                self.context, id=id_, uuid=temp_uuid)
            federation_list.append(federation.uuid)

        response = self.get_json('/federations')
        self.assertEqual(len(federation_list), len(response['federations']))
        uuids = [f['uuid'] for f in response['federations']]
        self.assertEqual(sorted(federation_list), sorted(uuids))

    def test_links(self):
        uuid = uuidutils.generate_uuid()
        obj_utils.create_test_federation(self.context, id=1, uuid=uuid)
        response = self.get_json('/federations/%s' % uuid)
        self.assertIn('links', response.keys())
        self.assertEqual(2, len(response['links']))
        self.assertIn(uuid, response['links'][0]['href'])
        for link in response['links']:
            bookmark = link['rel'] == 'bookmark'
            self.assertTrue(self.validate_link(link['href'],
                                               bookmark=bookmark))

    def test_collection_links(self):
        for id_ in range(5):
            obj_utils.create_test_federation(self.context, id=id_,
                                             uuid=uuidutils.generate_uuid())
        response = self.get_json('/federations/?limit=3')
        next_marker = response['federations'][-1]['uuid']
        self.assertIn(next_marker, response['next'])

    def test_collection_links_default_limit(self):
        cfg.CONF.set_override('max_limit', 3, 'api')
        for id_ in range(5):
            obj_utils.create_test_federation(self.context, id=id_,
                                             uuid=uuidutils.generate_uuid())
        response = self.get_json('/federations')
        self.assertEqual(3, len(response['federations']))

        next_marker = response['federations'][-1]['uuid']
        self.assertIn(next_marker, response['next'])


class TestPatch(api_base.FunctionalTest):
    def setUp(self):
        super(TestPatch, self).setUp()
        p = mock.patch.object(rpcapi.API, 'federation_update_async')
        self.mock_federation_update = p.start()
        self.mock_federation_update.side_effect = \
            self._sim_rpc_federation_update
        self.addCleanup(p.stop)

    def _sim_rpc_federation_update(self, federation, rollback=False):
        federation.save()
        return federation

    def test_member_join(self):
        f = obj_utils.create_test_federation(
            self.context, name='federation-example',
            uuid=uuidutils.generate_uuid(), member_ids=[])
        new_member = obj_utils.create_test_cluster(self.context)

        response = self.patch_json(
            '/federations/%s' % f.uuid,
            [{'path': '/member_ids', 'value': new_member.uuid, 'op': 'add'}])
        self.assertEqual(202, response.status_int)

        # make sure it was added:
        fed = self.get_json('/federations/%s' % f.uuid)
        self.assertTrue(new_member.uuid in fed['member_ids'])

    def test_member_unjoin(self):
        member = obj_utils.create_test_cluster(self.context)
        federation = obj_utils.create_test_federation(
            self.context, name='federation-example',
            uuid=uuidutils.generate_uuid(), member_ids=[member.uuid])

        response = self.patch_json(
            '/federations/%s' % federation.uuid,
            [{'path': '/member_ids', 'value': member.uuid, 'op': 'remove'}])
        self.assertEqual(202, response.status_int)

        # make sure it was deleted:
        fed = self.get_json('/federations/%s' % federation.uuid)
        self.assertFalse(member.uuid in fed['member_ids'])

    def test_join_non_existent_cluster(self):
        foo_uuid = uuidutils.generate_uuid()
        f = obj_utils.create_test_federation(
            self.context, name='federation-example',
            uuid=uuidutils.generate_uuid(), member_ids=[])

        response = self.patch_json(
            '/federations/%s' % f.uuid,
            [{'path': '/member_ids', 'value': foo_uuid, 'op': 'add'}],
            expect_errors=True)
        self.assertEqual(404, response.status_int)

    def test_unjoin_non_existent_cluster(self):
        foo_uuid = uuidutils.generate_uuid()
        f = obj_utils.create_test_federation(
            self.context, name='federation-example',
            uuid=uuidutils.generate_uuid(), member_ids=[])

        response = self.patch_json(
            '/federations/%s' % f.uuid,
            [{'path': '/member_ids', 'value': foo_uuid, 'op': 'remove'}],
            expect_errors=True)
        self.assertEqual(404, response.status_int)

    def test_join_cluster_already_member(self):
        cluster = obj_utils.create_test_cluster(self.context)
        f = obj_utils.create_test_federation(
            self.context, name='federation-example',
            uuid=uuidutils.generate_uuid(), member_ids=[cluster.uuid])

        response = self.patch_json(
            '/federations/%s' % f.uuid,
            [{'path': '/member_ids', 'value': cluster.uuid, 'op': 'add'}],
            expect_errors=True)
        self.assertEqual(409, response.status_int)

    def test_unjoin_non_member_cluster(self):
        cluster = obj_utils.create_test_cluster(self.context)
        f = obj_utils.create_test_federation(
            self.context, name='federation-example',
            uuid=uuidutils.generate_uuid(), member_ids=[])

        response = self.patch_json(
            '/federations/%s' % f.uuid,
            [{'path': '/member_ids', 'value': cluster.uuid, 'op': 'remove'}],
            expect_errors=True)
        self.assertEqual(404, response.status_int)


class TestPost(api_base.FunctionalTest):
    def setUp(self):
        super(TestPost, self).setUp()
        p = mock.patch.object(rpcapi.API, 'federation_create_async')
        self.mock_fed_create = p.start()
        self.mock_fed_create.side_effect = self._simulate_federation_create
        self.addCleanup(p.stop)
        self.hostcluster = obj_utils.create_test_cluster(self.context)

    def _simulate_federation_create(self, federation, create_timeout):
        federation.create()
        return federation

    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_create_federation(self, mock_utcnow):
        bdict = apiutils.federation_post_data(
            uuid=uuidutils.generate_uuid(),
            hostcluster_id=self.hostcluster.uuid)
        test_time = datetime.datetime(2000, 1, 1, 0, 0)
        mock_utcnow.return_value = test_time

        response = self.post_json('/federations', bdict)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(202, response.status_int)
        self.assertTrue(uuidutils.is_uuid_like(response.json['uuid']))

    def test_create_federation_no_hostcluster_id(self):
        bdict = apiutils.federation_post_data(uuid=uuidutils.generate_uuid())
        del bdict['hostcluster_id']
        response = self.post_json('/federations', bdict, expect_errors=True)
        self.assertEqual(400, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_create_federation_hostcluster_does_not_exist(self):
        bdict = apiutils.federation_post_data(
            uuid=uuidutils.generate_uuid(),
            hostcluster_id=uuidutils.generate_uuid())
        response = self.post_json('/federations', bdict, expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_create_federation_no_dns_zone_name(self):
        bdict = apiutils.federation_post_data(
            uuid=uuidutils.generate_uuid(),
            hostcluster_id=self.hostcluster.uuid)
        del bdict['properties']
        response = self.post_json('/federations', bdict, expect_errors=True)
        self.assertEqual(400, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_create_federation_generate_uuid(self):
        bdict = apiutils.federation_post_data(
            hostcluster_id=self.hostcluster.uuid)
        del bdict['uuid']
        response = self.post_json('/federations', bdict)
        self.assertEqual(202, response.status_int)

    def test_create_federation_with_invalid_name(self):
        invalid_names = [
            'x' * 243, '123456', '123456test_federation',
            '-test_federation', '.test_federation', '_test_federation', ''
        ]

        for value in invalid_names:
            bdict = apiutils.federation_post_data(
                uuid=uuidutils.generate_uuid(), name=value,
                hostcluster_id=self.hostcluster.uuid)
            response = self.post_json('/federations', bdict,
                                      expect_errors=True)
            self.assertEqual('application/json', response.content_type)
            self.assertEqual(400, response.status_int)
            self.assertTrue(response.json['errors'])

    def test_create_federation_with_valid_name(self):
        valid_names = [
            'test_federation123456', 'test-federation', 'test.federation',
            'testfederation.', 'testfederation-', 'testfederation_',
            'test.-_federation', 'Testfederation'
        ]

        for value in valid_names:
            bdict = apiutils.federation_post_data(
                name=value, hostcluster_id=self.hostcluster.uuid)
            bdict['uuid'] = uuidutils.generate_uuid()
            response = self.post_json('/federations', bdict)
            self.assertEqual(202, response.status_int)

    def test_create_federation_without_name(self):
        bdict = apiutils.federation_post_data(
            uuid=uuidutils.generate_uuid(),
            hostcluster_id=self.hostcluster.uuid)
        del bdict['name']
        response = self.post_json('/federations', bdict)
        self.assertEqual(202, response.status_int)


class TestDelete(api_base.FunctionalTest):
    def setUp(self):
        super(TestDelete, self).setUp()
        self.federation = obj_utils.create_test_federation(
            self.context, name='federation-example',
            uuid=uuidutils.generate_uuid())
        p = mock.patch.object(rpcapi.API, 'federation_delete_async')
        self.mock_federation_delete = p.start()
        self.mock_federation_delete.side_effect = \
            self._simulate_federation_delete
        self.addCleanup(p.stop)

    def _simulate_federation_delete(self, federation_uuid):
        federation = objects.Federation.get_by_uuid(self.context,
                                                    federation_uuid)
        federation.destroy()

    def test_delete_federation(self):
        self.delete('/federations/%s' % self.federation.uuid)
        response = self.get_json('/federations/%s' % self.federation.uuid,
                                 expect_errors=True)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['errors'])

    def test_delete_federation_not_found(self):
        delete = self.delete('/federations/%s' % uuidutils.generate_uuid(),
                             expect_errors=True)
        self.assertEqual(404, delete.status_int)
        self.assertEqual('application/json', delete.content_type)
        self.assertTrue(delete.json['errors'])

    def test_delete_federation_with_name(self):
        delete = self.delete('/federations/%s' % self.federation.name)
        self.assertEqual(204, delete.status_int)

    def test_delete_federation_with_name_not_found(self):
        delete = self.delete('/federations/%s' % 'foo',
                             expect_errors=True)
        self.assertEqual(404, delete.status_int)
        self.assertEqual('application/json', delete.content_type)
        self.assertTrue(delete.json['errors'])



# Copyright 2015 NEC Corporation.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from taskflow import engines
from taskflow.patterns import linear_flow
from unittest import mock

from magnum.conductor.tasks import heat_tasks
from magnum.tests import base


class HeatTasksTests(base.TestCase):

    def setUp(self):
        super(HeatTasksTests, self).setUp()
        self.heat_client = mock.MagicMock(name='heat_client')

    def _get_create_stack_flow(self, heat_client):
        flow = linear_flow.Flow("create stack flow")
        flow.add(
            heat_tasks.CreateStack(
                os_client=heat_client,
                requires=('stack_name', 'parameters', 'template', 'files'),
                provides='new_stack',
            ),
        )
        return flow

    def _get_update_stack_flow(self, heat_client):
        flow = linear_flow.Flow("update stack flow")
        flow.add(
            heat_tasks.UpdateStack(
                os_client=heat_client,
                requires=('stack_id', 'parameters', 'template', 'files'),
            ),
        )
        return flow

    def _get_delete_stack_flow(self, heat_client):
        flow = linear_flow.Flow("delete stack flow")
        flow.add(
            heat_tasks.DeleteStack(
                os_client=heat_client,
                requires=('stack_id'),
            ),
        )
        return flow

    def test_create_stack(self):
        heat_client = mock.MagicMock(name='heat_client')
        stack_id = 'stack_id'
        stack_name = 'stack_name'
        stack = {
            'stack': {
                'id': stack_id
            }
        }
        heat_client.stacks.create.return_value = stack
        flow_store = {
            'stack_name': stack_name,
            'parameters': 'parameters',
            'template': 'template',
            'files': 'files'
        }
        flow = self._get_create_stack_flow(heat_client)

        result = engines.run(flow, store=flow_store)
        heat_client.stacks.create.assert_called_once_with(**flow_store)
        self.assertEqual(stack_id, result['new_stack']['stack']['id'])

    def test_create_stack_with_error(self):
        heat_client = mock.MagicMock(name='heat_client')
        heat_client.stacks.create.side_effect = ValueError
        stack_name = 'stack_name'
        flow_store = {
            'stack_name': stack_name,
            'parameters': 'parameters',
            'template': 'template',
            'files': 'files'
        }
        flow = self._get_create_stack_flow(heat_client)

        self.assertRaises(ValueError, engines.run, flow, store=flow_store)

    def test_update_stack(self):
        heat_client = mock.MagicMock(name='heat_client')
        stack_id = 'stack_id'
        flow_store = {
            'stack_id': stack_id,
            'parameters': 'parameters',
            'template': 'template',
            'files': 'files'
        }
        flow = self._get_update_stack_flow(heat_client)
        expected_params = dict(flow_store)
        del expected_params['stack_id']

        engines.run(flow, store=flow_store)
        heat_client.stacks.update.assert_called_once_with(stack_id,
                                                          **expected_params)

    def test_update_stack_with_error(self):
        heat_client = mock.MagicMock(name='heat_client')
        heat_client.stacks.update.side_effect = ValueError
        stack_id = 'stack_id'
        flow_store = {
            'stack_id': stack_id,
            'parameters': 'parameters',
            'template': 'template',
            'files': 'files'
        }
        flow = self._get_update_stack_flow(heat_client)

        self.assertRaises(ValueError, engines.run, flow, store=flow_store)

    def test_delete_stack(self):
        heat_client = mock.MagicMock(name='heat_client')
        stack_id = 'stack_id'
        flow_store = {'stack_id': stack_id}
        flow = self._get_delete_stack_flow(heat_client)

        engines.run(flow, store=flow_store)
        heat_client.stacks.delete.assert_called_once_with(stack_id)

    def test_delete_stack_with_error(self):
        heat_client = mock.MagicMock(name='heat_client')
        heat_client.stacks.delete.side_effect = ValueError
        stack_id = 'stack_id'
        flow_store = {'stack_id': stack_id}
        flow = self._get_delete_stack_flow(heat_client)

        self.assertRaises(ValueError, engines.run, flow, store=flow_store)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import collections
import os
from pbr.version import SemanticVersion as SV
import six

from string import ascii_letters
from string import digits

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

from heatclient.common import template_utils
from heatclient import exc as heatexc

from magnum.common import clients
from magnum.common import context as mag_ctx
from magnum.common import exception
from magnum.common import keystone
from magnum.common import octavia
from magnum.common import short_id
from magnum.common.x509 import operations as x509
from magnum.conductor.handlers.common import cert_manager
from magnum.conductor.handlers.common import trust_manager
from magnum.conductor import utils as conductor_utils
from magnum.drivers.common import driver
from magnum.drivers.common import k8s_monitor
from magnum.i18n import _
from magnum.objects import fields

LOG = logging.getLogger(__name__)


NodeGroupStatus = collections.namedtuple('NodeGroupStatus',
                                         'name status reason is_default')


@six.add_metaclass(abc.ABCMeta)
class HeatDriver(driver.Driver):
    """Base Driver class for using Heat

       Abstract class for implementing Drivers that leverage OpenStack Heat for
       orchestrating cluster lifecycle operations
    """

    def _extract_template_definition_up(self, context, cluster,
                                        cluster_template,
                                        scale_manager=None):
        ct_obj = conductor_utils.retrieve_ct_by_name_or_uuid(
            context,
            cluster_template)
        definition = self.get_template_definition()
        return definition.extract_definition(context, ct_obj,
                                             cluster,
                                             scale_manager=scale_manager)

    def _extract_template_definition(self, context, cluster,
                                     scale_manager=None,
                                     nodegroups=None):
        cluster_template = conductor_utils.retrieve_cluster_template(context,
                                                                     cluster)
        definition = self.get_template_definition()
        return definition.extract_definition(context, cluster_template,
                                             cluster,
                                             nodegroups=nodegroups,
                                             scale_manager=scale_manager)

    def _get_env_files(self, template_path, env_rel_paths):
        template_dir = os.path.dirname(template_path)
        env_abs_paths = [os.path.join(template_dir, f) for f in env_rel_paths]
        environment_files = []
        env_map, merged_env = (
            template_utils.process_multiple_environments_and_files(
                env_paths=env_abs_paths, env_list_tracker=environment_files))
        return environment_files, env_map

    @abc.abstractmethod
    def get_template_definition(self):
        """return an implementation of

           magnum.drivers.common.drivers.heat.TemplateDefinition
        """

        raise NotImplementedError("Must implement 'get_template_definition'")

    def create_federation(self, context, federation):
        return NotImplementedError("Must implement 'create_federation'")

    def update_federation(self, context, federation):
        return NotImplementedError("Must implement 'update_federation'")

    def delete_federation(self, context, federation):
        return NotImplementedError("Must implement 'delete_federation'")

    def update_nodegroup(self, context, cluster, nodegroup):
        # we just need to save the nodegroup here. This is because,
        # at the moment, this method is used to update min and max node
        # counts.
        nodegroup.save()

    def delete_nodegroup(self, context, cluster, nodegroup):
        # Default nodegroups share stack_id so it will be deleted
        # as soon as the cluster gets destroyed
        if not nodegroup.stack_id:
            nodegroup.destroy()
        else:
            osc = clients.OpenStackClients(context)
            self._delete_stack(context, osc, nodegroup.stack_id)

    def update_cluster_status(self, context, cluster, use_admin_ctx=False):
        if cluster.stack_id is None:
            # NOTE(mgoddard): During cluster creation it is possible to poll
            # the cluster before its heat stack has been created. See bug
            # 1682058.
            return
        if use_admin_ctx:
            stack_ctx = context
        else:
            stack_ctx = mag_ctx.make_cluster_context(cluster)
        poller = HeatPoller(clients.OpenStackClients(stack_ctx), context,
                            cluster, self)
        poller.poll_and_check()

    def create_cluster(self, context, cluster, cluster_create_timeout):
        stack = self._create_stack(context, clients.OpenStackClients(context),
                                   cluster, cluster_create_timeout)
        # TODO(randall): keeping this for now to reduce/eliminate data
        # migration. Should probably come up with something more generic in
        # the future once actual non-heat-based drivers are implemented.
        cluster.stack_id = stack['stack']['id']

    def update_cluster(self, context, cluster, scale_manager=None,
                       rollback=False):
        self._update_stack(context, cluster, scale_manager, rollback)

    def create_nodegroup(self, context, cluster, nodegroup):
        stack = self._create_stack(context, clients.OpenStackClients(context),
                                   cluster, cluster.create_timeout,
                                   nodegroup=nodegroup)
        nodegroup.stack_id = stack['stack']['id']

    def get_nodegroup_extra_params(self, cluster, osc):
        raise NotImplementedError("Must implement "
                                  "'get_nodegroup_extra_params'")

    @abc.abstractmethod
    def upgrade_cluster(self, context, cluster, cluster_template,
                        max_batch_size, nodegroup, scale_manager=None,
                        rollback=False):
        raise NotImplementedError("Must implement 'upgrade_cluster'")

    def delete_cluster(self, context, cluster):
        self.pre_delete_cluster(context, cluster)

        LOG.info("Starting to delete cluster %s", cluster.uuid)
        osc = clients.OpenStackClients(context)
        for ng in cluster.nodegroups:
            ng.status = fields.ClusterStatus.DELETE_IN_PROGRESS
            ng.save()
            if ng.is_default:
                continue
            self._delete_stack(context, osc, ng.stack_id)
        self._delete_stack(context, osc, cluster.default_ng_master.stack_id)

    def resize_cluster(self, context, cluster, resize_manager,
                       node_count, nodes_to_remove, nodegroup=None,
                       rollback=False):
        self._resize_stack(context, cluster, resize_manager,
                           node_count, nodes_to_remove, nodegroup=nodegroup,
                           rollback=rollback)

    def _create_stack(self, context, osc, cluster, cluster_create_timeout,
                      nodegroup=None):

        nodegroups = [nodegroup] if nodegroup else None
        template_path, heat_params, env_files = (
            self._extract_template_definition(context, cluster,
                                              nodegroups=nodegroups))

        tpl_files, template = template_utils.get_template_contents(
            template_path)

        environment_files, env_map = self._get_env_files(template_path,
                                                         env_files)
        tpl_files.update(env_map)

        # Make sure we end up with a valid hostname
        valid_chars = set(ascii_letters + digits + '-')

        # valid hostnames are 63 chars long, leaving enough room
        # to add the random id (for uniqueness)
        if nodegroup is None:
            stack_name = cluster.name[:30]
        else:
            stack_name = "%s-%s" % (cluster.name[:20], nodegroup.name[:9])
        stack_name = stack_name.replace('_', '-')
        stack_name = stack_name.replace('.', '-')
        stack_name = ''.join(filter(valid_chars.__contains__, stack_name))

        # Make sure no duplicate stack name
        stack_name = '%s-%s' % (stack_name, short_id.generate_id())
        stack_name = stack_name.lower()
        if cluster_create_timeout:
            heat_timeout = cluster_create_timeout
        else:
            # no cluster_create_timeout value was passed in to the request
            # so falling back on configuration file value
            heat_timeout = cfg.CONF.cluster_heat.create_timeout

        heat_params['is_cluster_stack'] = nodegroup is None

        if nodegroup:
            # In case we are creating a new stack for a new nodegroup then
            # we need to extract more params.
            heat_params.update(self.get_nodegroup_extra_params(cluster, osc))

        fields = {
            'stack_name': stack_name,
            'parameters': heat_params,
            'environment_files': environment_files,
            'template': template,
            'files': tpl_files,
            'timeout_mins': heat_timeout
        }
        created_stack = osc.heat().stacks.create(**fields)

        return created_stack

    def _update_stack(self, context, cluster, scale_manager=None,
                      rollback=False):
        # update worked properly only for scaling nodes up and down
        # before nodegroups. Maintain this logic until we deprecate
        # and remove the command.
        # Fixed behaviour Id84e5d878b21c908021e631514c2c58b3fe8b8b0
        nodegroup = cluster.default_ng_worker
        definition = self.get_template_definition()
        scale_params = definition.get_scale_params(
            context,
            cluster,
            nodegroup.node_count,
            scale_manager,
            nodes_to_remove=None)

        fields = {
            'parameters': scale_params,
            'existing': True,
            'disable_rollback': not rollback
        }

        LOG.info('Updating cluster %s stack %s with these params: %s',
                 cluster.uuid, nodegroup.stack_id, scale_params)
        osc = clients.OpenStackClients(context)
        osc.heat().stacks.update(nodegroup.stack_id, **fields)

    def _resize_stack(self, context, cluster, resize_manager,
                      node_count, nodes_to_remove, nodegroup=None,
                      rollback=False):
        definition = self.get_template_definition()
        scale_params = definition.get_scale_params(
            context,
            cluster,
            nodegroup.node_count,
            resize_manager,
            nodes_to_remove=nodes_to_remove)

        fields = {
            'parameters': scale_params,
            'existing': True,
            'disable_rollback': not rollback
        }

        LOG.info('Resizing cluster %s stack %s with these params: %s',
                 cluster.uuid, nodegroup.stack_id, scale_params)
        osc = clients.OpenStackClients(context)
        osc.heat().stacks.update(nodegroup.stack_id, **fields)

    def _delete_stack(self, context, osc, stack_id):
        osc.heat().stacks.delete(stack_id)


class KubernetesDriver(HeatDriver):
    """Base driver for Kubernetes clusters."""

    def get_monitor(self, context, cluster):
        return k8s_monitor.K8sMonitor(context, cluster)

    def get_scale_manager(self, context, osclient, cluster):
        # FIXME: Until the kubernetes client is fixed, remove
        # the scale_manager.
        # https://bugs.launchpad.net/magnum/+bug/1746510
        return None

    def pre_delete_cluster(self, context, cluster):
        """Delete cloud resources before deleting the cluster."""
        if keystone.is_octavia_enabled():
            LOG.info("Starting to delete loadbalancers for cluster %s",
                     cluster.uuid)
            octavia.delete_loadbalancers(context, cluster)

    def upgrade_cluster(self, context, cluster, cluster_template,
                        max_batch_size, nodegroup, scale_manager=None,
                        rollback=False):
        raise NotImplementedError("Must implement 'upgrade_cluster'")


class FedoraKubernetesDriver(KubernetesDriver):
    """Base driver for Kubernetes clusters."""

    def get_heat_params(self, cluster_template):
        heat_params = {}
        try:
            kube_tag = cluster_template.labels["kube_tag"]
            kube_tag_params = {
                "kube_tag": kube_tag,
                "kube_version": kube_tag,
                "master_kube_tag": kube_tag,
                "minion_kube_tag": kube_tag,
            }
            heat_params.update(kube_tag_params)
        except KeyError:
            LOG.debug(("Cluster template %s does not contain a "
                       "valid kube_tag"), cluster_template.name)

        # If both keys are present, only ostree_commit is chosen.
        for ostree_tag in ["ostree_commit", "ostree_remote"]:
            if ostree_tag not in cluster_template.labels:
                continue
            try:
                ostree_param = {
                    ostree_tag: cluster_template.labels[ostree_tag]
                }
                heat_params.update(ostree_param)
                break
            except KeyError:
                LOG.debug("Cluster template %s does not define %s",
                          cluster_template.name, ostree_tag)

        upgrade_labels = ['kube_tag', 'ostree_remote', 'ostree_commit']
        if not any([u in heat_params.keys() for u in upgrade_labels]):
            reason = ("Cluster template %s does not contain any supported "
                      "upgrade labels: [%s]") % (cluster_template.name,
                                                 ', '.join(upgrade_labels))
            raise exception.InvalidClusterTemplateForUpgrade(reason=reason)

        return heat_params

    @staticmethod
    def get_new_labels(nodegroup, cluster_template):
        new_labels = nodegroup.labels.copy()
        if 'kube_tag' in cluster_template.labels:
            new_kube_tag = cluster_template.labels['kube_tag']
            new_labels.update({'kube_tag': new_kube_tag})
        return new_labels

    def upgrade_cluster(self, context, cluster, cluster_template,  # noqa: C901
                        max_batch_size, nodegroup, scale_manager=None,
                        rollback=False):
        osc = clients.OpenStackClients(context)

        # Use this just to check that we are not downgrading.
        heat_params = {
            "update_max_batch_size": max_batch_size,
        }

        if 'kube_tag' in nodegroup.labels:
            heat_params['kube_tag'] = nodegroup.labels['kube_tag']

        current_addons = {}
        new_addons = {}
        for label in cluster_template.labels:
            # This is upgrade API, so we don't introduce new stuff by this API,
            # but just focus on the version change.
            new_addons[label] = cluster_template.labels[label]
            if ((label.endswith('_tag') or
                 label.endswith('_version')) and label in heat_params):
                current_addons[label] = heat_params[label]
                try:
                    if (SV.from_pip_string(new_addons[label]) <
                            SV.from_pip_string(current_addons[label])):
                        raise exception.InvalidVersion(tag=label)
                except exception.InvalidVersion:
                    raise
                except Exception as e:
                    # NOTE(flwang): Different cloud providers may use different
                    # tag/version format which maybe not able to parse by
                    # SemanticVersion. For this case, let's just skip it.
                    LOG.debug("Failed to parse tag/version %s", str(e))

        # Since the above check passed just
        # hardcode what we want to send to heat.
        # Rules: 1. No downgrade 2. Explicitly override 3. Merging based on set
        # Update heat_params based on the data generated above
        heat_params.update(self.get_heat_params(cluster_template))

        stack_id = nodegroup.stack_id
        if nodegroup is not None and not nodegroup.is_default:
            heat_params['is_cluster_stack'] = False
            # For now set the worker_role explicitly in order to
            # make sure that the is_master condition fails.
            heat_params['worker_role'] = nodegroup.role

        # we need to set the whole dict to the object
        # and not just update the existing labels. This
        # is how obj_what_changed works.
        nodegroup.labels = new_labels = self.get_new_labels(nodegroup,
                                                            cluster_template)

        if nodegroup.is_default:
            cluster.cluster_template_id = cluster_template.uuid
            cluster.labels = new_labels
            if nodegroup.role == 'master':
                other_default_ng = cluster.default_ng_worker
            else:
                other_default_ng = cluster.default_ng_master
            other_default_ng.labels = new_labels
            other_default_ng.save()

        fields = {
            'existing': True,
            'parameters': heat_params,
            'disable_rollback': not rollback
        }
        LOG.info('Upgrading cluster %s stack %s with these params: %s',
                 cluster.uuid, nodegroup.stack_id, heat_params)
        osc.heat().stacks.update(stack_id, **fields)

    def get_nodegroup_extra_params(self, cluster, osc):
        network = osc.heat().resources.get(cluster.stack_id, 'network')
        secgroup = osc.heat().resources.get(cluster.stack_id,
                                            'secgroup_kube_minion')
        for output in osc.heat().stacks.get(cluster.stack_id).outputs:
            if output['output_key'] == 'api_address':
                api_address = output['output_value']
                break
        extra_params = {
            'existing_master_private_ip': api_address,
            'existing_security_group': secgroup.attributes['id'],
            'fixed_network': network.attributes['fixed_network'],
            'fixed_subnet': network.attributes['fixed_subnet'],
        }
        return extra_params

    def rotate_ca_certificate(self, context, cluster):
        cluster_template = conductor_utils.retrieve_cluster_template(context,
                                                                     cluster)
        if cluster_template.cluster_distro not in ["fedora-coreos"]:
            raise exception.NotSupported("Rotating the CA certificate is "
                                         "not supported for cluster with "
                                         "cluster_distro: %s." %
                                         cluster_template.cluster_distro)
        osc = clients.OpenStackClients(context)
        rollback = True
        heat_params = {}

        csr_keys = x509.generate_csr_and_key(u"Kubernetes Service Account")

        heat_params['kube_service_account_key'] = \
            csr_keys["public_key"].replace("\n", "\\n")
        heat_params['kube_service_account_private_key'] = \
            csr_keys["private_key"].replace("\n", "\\n")

        fields = {
            'existing': True,
            'parameters': heat_params,
            'disable_rollback': not rollback
        }
        osc.heat().stacks.update(cluster.stack_id, **fields)


class HeatPoller(object):

    def __init__(self, openstack_client, context, cluster, cluster_driver):
        self.openstack_client = openstack_client
        self.context = context
        self.cluster = cluster
        self.cluster_template = conductor_utils.retrieve_cluster_template(
            self.context, cluster)
        self.template_def = cluster_driver.get_template_definition()

    def poll_and_check(self):
        # TODO(yuanying): temporary implementation to update api_address,
        # node_addresses and cluster status
        ng_statuses = list()
        self.default_ngs = list()
        for nodegroup in self.cluster.nodegroups:
            self.nodegroup = nodegroup
            if self.nodegroup.is_default:
                self.default_ngs.append(self.nodegroup)
            status = self.extract_nodegroup_status()
            # In case a non-default nodegroup is deleted, None
            # is returned. We shouldn't add None in the list
            if status is not None:
                ng_statuses.append(status)
        self.aggregate_nodegroup_statuses(ng_statuses)

    def extract_nodegroup_status(self):

        if self.nodegroup.stack_id is None:
            # There is a slight window for a race condition here. If
            # a nodegroup is created and just before the stack_id is
            # assigned to it, this periodic task is executed, the
            # periodic task would try to find the status of the
            # stack with id = None. At that time the nodegroup status
            # is already set to CREATE_IN_PROGRESS by the conductor.
            # Keep this status for this loop until the stack_id is assigned.
            return NodeGroupStatus(name=self.nodegroup.name,
                                   status=self.nodegroup.status,
                                   is_default=self.nodegroup.is_default,
                                   reason=self.nodegroup.status_reason)

        try:
            # Do not resolve outputs by default. Resolving all
            # node IPs is expensive on heat.
            stack = self.openstack_client.heat().stacks.get(
                self.nodegroup.stack_id, resolve_outputs=False)

            if stack.stack_status in (fields.ClusterStatus.CREATE_COMPLETE,
                                      fields.ClusterStatus.UPDATE_COMPLETE):
                # Resolve all outputs if the stack is COMPLETE
                stack = self.openstack_client.heat().stacks.get(
                    self.nodegroup.stack_id, resolve_outputs=True)

                self._sync_cluster_and_template_status(stack)
            elif stack.stack_status != self.nodegroup.status:
                self.template_def.nodegroup_output_mappings = list()
                self.template_def.update_outputs(
                    stack, self.cluster_template, self.cluster,
                    nodegroups=[self.nodegroup])
                self._sync_cluster_status(stack)

            # poll_and_check is detached and polling long time to check
            # status, so another user/client can call delete cluster/stack.
            if stack.stack_status == fields.ClusterStatus.DELETE_COMPLETE:
                if self.nodegroup.is_default:
                    self._check_delete_complete()
                else:
                    self.nodegroup.destroy()
                    return

            if stack.stack_status in (fields.ClusterStatus.CREATE_FAILED,
                                      fields.ClusterStatus.DELETE_FAILED,
                                      fields.ClusterStatus.UPDATE_FAILED,
                                      fields.ClusterStatus.ROLLBACK_COMPLETE,
                                      fields.ClusterStatus.ROLLBACK_FAILED):
                self._sync_cluster_and_template_status(stack)
                self._nodegroup_failed(stack)
        except heatexc.NotFound:
            self._sync_missing_heat_stack()
        return NodeGroupStatus(name=self.nodegroup.name,
                               status=self.nodegroup.status,
                               is_default=self.nodegroup.is_default,
                               reason=self.nodegroup.status_reason)

    def aggregate_nodegroup_statuses(self, ng_statuses):
        # NOTE(ttsiouts): Aggregate the nodegroup statuses and set the
        # cluster overall status.
        FAILED = '_FAILED'
        IN_PROGRESS = '_IN_PROGRESS'
        COMPLETE = '_COMPLETE'
        UPDATE = 'UPDATE'
        DELETE = 'DELETE'

        previous_state = self.cluster.status
        self.cluster.status_reason = None

        non_default_ngs_exist = any(not ns.is_default for ns in ng_statuses)
        # Both default nodegroups will have the same status so it's
        # enough to check one of them.
        default_ng_status = self.cluster.default_ng_master.status
        # Whatever action is going on in a cluster that has
        # non-default ngs, we call it update except for delete.
        action = DELETE if default_ng_status.startswith(DELETE) else UPDATE
        # Keep priority to the states below
        for state in (IN_PROGRESS, FAILED, COMPLETE):
            if any(ns.status.endswith(state) for ns in ng_statuses):
                if non_default_ngs_exist:
                    status = getattr(fields.ClusterStatus, action+state)
                else:
                    # If there are no non-default NGs
                    # just use the default NG's status.
                    status = default_ng_status
                self.cluster.status = status
                break

        if self.cluster.status == fields.ClusterStatus.CREATE_COMPLETE:
            # Consider the scenario where the user:
            # - creates the cluster (cluster: create_complete)
            # - adds a nodegroup (cluster: update_complete)
            # - deletes the nodegroup
            # The cluster should go to CREATE_COMPLETE only if the previous
            # state was CREATE_COMPLETE or CREATE_IN_PROGRESS. In all other
            # cases, just go to UPDATE_COMPLETE.
            if previous_state not in (fields.ClusterStatus.CREATE_COMPLETE,
                                      fields.ClusterStatus.CREATE_IN_PROGRESS):
                self.cluster.status = fields.ClusterStatus.UPDATE_COMPLETE

        # Summarize the failed reasons.
        if self.cluster.status.endswith(FAILED):
            reasons = ["%s failed" % (ns.name)
                       for ns in ng_statuses
                       if ns.status.endswith(FAILED)]
            self.cluster.status_reason = ', '.join(reasons)

        self.cluster.save()

    def _delete_complete(self):
        LOG.info('Cluster has been deleted, stack_id: %s',
                 self.cluster.stack_id)
        try:
            trust_manager.delete_trustee_and_trust(self.openstack_client,
                                                   self.context,
                                                   self.cluster)
            cert_manager.delete_certificates_from_cluster(self.cluster,
                                                          context=self.context)
            cert_manager.delete_client_files(self.cluster,
                                             context=self.context)

        except exception.ClusterNotFound:
            LOG.info('The cluster %s has been deleted by others.',
                     self.cluster.uuid)

    def _sync_cluster_status(self, stack):
        self.nodegroup.status = stack.stack_status
        self.nodegroup.status_reason = stack.stack_status_reason
        self.nodegroup.save()

    def get_version_info(self, stack):
        stack_param = self.template_def.get_heat_param(
            cluster_attr='coe_version')
        if stack_param:
            self.cluster.coe_version = stack.parameters[stack_param]

        version_module_path = self.template_def.driver_module_path+'.version'
        try:
            ver = importutils.import_module(version_module_path)
            container_version = ver.container_version
        except Exception:
            container_version = None
        self.cluster.container_version = container_version

    def _sync_cluster_and_template_status(self, stack):
        self.template_def.nodegroup_output_mappings = list()
        self.template_def.update_outputs(stack, self.cluster_template,
                                         self.cluster,
                                         nodegroups=[self.nodegroup])
        self.get_version_info(stack)
        self._sync_cluster_status(stack)

    def _nodegroup_failed(self, stack):
        LOG.error('Nodegroup error, stack status: %(ng_status)s, '
                  'stack_id: %(stack_id)s, '
                  'reason: %(reason)s',
                  {'ng_status': stack.stack_status,
                   'stack_id': self.nodegroup.stack_id,
                   'reason': self.nodegroup.status_reason})

    def _sync_missing_heat_stack(self):
        if self.nodegroup.status == fields.ClusterStatus.DELETE_IN_PROGRESS:
            self._sync_missing_stack(fields.ClusterStatus.DELETE_COMPLETE)
            if self.nodegroup.is_default:
                self._check_delete_complete()
        elif self.nodegroup.status == fields.ClusterStatus.CREATE_IN_PROGRESS:
            self._sync_missing_stack(fields.ClusterStatus.CREATE_FAILED)
        elif self.nodegroup.status == fields.ClusterStatus.UPDATE_IN_PROGRESS:
            self._sync_missing_stack(fields.ClusterStatus.UPDATE_FAILED)

    def _check_delete_complete(self):
        default_ng_statuses = [ng.status for ng in self.default_ngs]
        if all(status == fields.ClusterStatus.DELETE_COMPLETE
               for status in default_ng_statuses):
            self._delete_complete()

    def _sync_missing_stack(self, new_status):
        self.nodegroup.status = new_status
        self.nodegroup.status_reason = _("Stack with id %s not found in "
                                         "Heat.") % self.cluster.stack_id
        self.nodegroup.save()
        LOG.info("Nodegroup with id %(id)s has been set to "
                 "%(status)s due to stack with id %(sid)s "
                 "not found in Heat.",
                 {'id': self.nodegroup.uuid, 'status': self.nodegroup.status,
                  'sid': self.nodegroup.stack_id})



# Copyright (c) 2018 European Organization for Nuclear Research.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
from unittest.mock import patch

from heatclient import exc

from magnum.common import exception
from magnum.conductor.handlers import nodegroup_conductor
from magnum.objects import fields
from magnum.tests.unit.db import base as db_base
from magnum.tests.unit.objects import utils as obj_utils


class TestHandler(db_base.DbTestCase):

    def setUp(self):
        super(TestHandler, self).setUp()
        self.handler = nodegroup_conductor.Handler()
        self.cluster = obj_utils.create_test_cluster(self.context)
        self.nodegroup = obj_utils.create_test_nodegroup(
            self.context, cluster_id=self.cluster.uuid)

    @patch('magnum.drivers.common.driver.Driver.get_driver')
    def test_nodegroup_create(self, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_get_driver.return_value = mock_driver
        nodegroup = mock.MagicMock()
        self.handler.nodegroup_create(self.context, self.cluster, nodegroup)
        mock_driver.create_nodegroup.assert_called_once_with(self.context,
                                                             self.cluster,
                                                             nodegroup)
        nodegroup.create.assert_called_once()
        nodegroup.save.assert_called_once()
        self.assertEqual(fields.ClusterStatus.UPDATE_IN_PROGRESS,
                         self.cluster.status)
        self.assertEqual(fields.ClusterStatus.CREATE_IN_PROGRESS,
                         nodegroup.status)

    @patch('magnum.drivers.common.driver.Driver.get_driver')
    def test_nodegroup_create_failed(self, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_get_driver.return_value = mock_driver
        side_effect = NotImplementedError("Test failure")
        mock_driver.create_nodegroup.side_effect = side_effect
        nodegroup = mock.MagicMock()
        self.assertRaises(NotImplementedError, self.handler.nodegroup_create,
                          self.context, self.cluster, nodegroup)
        mock_driver.create_nodegroup.assert_called_once_with(self.context,
                                                             self.cluster,
                                                             nodegroup)
        nodegroup.create.assert_called_once()
        nodegroup.save.assert_called_once()
        self.assertEqual(fields.ClusterStatus.UPDATE_FAILED,
                         self.cluster.status)
        self.assertEqual(fields.ClusterStatus.CREATE_FAILED,
                         nodegroup.status)
        self.assertEqual("Test failure", nodegroup.status_reason)

    @patch('magnum.drivers.common.driver.Driver.get_driver')
    def test_nodegroup_create_failed_bad_request(self, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_get_driver.return_value = mock_driver
        side_effect = exc.HTTPBadRequest("Bad request")
        mock_driver.create_nodegroup.side_effect = side_effect
        nodegroup = mock.MagicMock()
        self.assertRaises(exception.InvalidParameterValue,
                          self.handler.nodegroup_create,
                          self.context, self.cluster, nodegroup)
        mock_driver.create_nodegroup.assert_called_once_with(self.context,
                                                             self.cluster,
                                                             nodegroup)
        nodegroup.create.assert_called_once()
        nodegroup.save.assert_called_once()
        self.assertEqual(fields.ClusterStatus.UPDATE_FAILED,
                         self.cluster.status)
        self.assertEqual(fields.ClusterStatus.CREATE_FAILED,
                         nodegroup.status)
        self.assertEqual("ERROR: Bad request", nodegroup.status_reason)

    @patch('magnum.drivers.common.driver.Driver.get_driver')
    def test_nodegroup_udpate(self, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_get_driver.return_value = mock_driver
        self.handler.nodegroup_update(self.context, self.cluster,
                                      self.nodegroup)
        mock_driver.update_nodegroup.assert_called_once_with(self.context,
                                                             self.cluster,
                                                             self.nodegroup)
        self.assertEqual(fields.ClusterStatus.UPDATE_IN_PROGRESS,
                         self.cluster.status)
        self.assertEqual(fields.ClusterStatus.UPDATE_IN_PROGRESS,
                         self.nodegroup.status)

    @patch('magnum.drivers.common.driver.Driver.get_driver')
    def test_nodegroup_update_failed(self, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_get_driver.return_value = mock_driver
        side_effect = NotImplementedError("Update failed")
        mock_driver.update_nodegroup.side_effect = side_effect
        self.assertRaises(NotImplementedError, self.handler.nodegroup_update,
                          self.context, self.cluster, self.nodegroup)
        mock_driver.update_nodegroup.assert_called_once_with(self.context,
                                                             self.cluster,
                                                             self.nodegroup)
        self.assertEqual(fields.ClusterStatus.UPDATE_FAILED,
                         self.cluster.status)
        self.assertEqual(fields.ClusterStatus.UPDATE_FAILED,
                         self.nodegroup.status)
        self.assertEqual("Update failed", self.nodegroup.status_reason)

    @patch('magnum.drivers.common.driver.Driver.get_driver')
    def test_nodegroup_update_failed_bad_request(self, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_get_driver.return_value = mock_driver
        side_effect = exc.HTTPBadRequest("Bad request")
        mock_driver.update_nodegroup.side_effect = side_effect
        self.assertRaises(exception.InvalidParameterValue,
                          self.handler.nodegroup_update,
                          self.context, self.cluster, self.nodegroup)
        mock_driver.update_nodegroup.assert_called_once_with(self.context,
                                                             self.cluster,
                                                             self.nodegroup)
        self.assertEqual(fields.ClusterStatus.UPDATE_FAILED,
                         self.cluster.status)
        self.assertEqual(fields.ClusterStatus.UPDATE_FAILED,
                         self.nodegroup.status)
        self.assertEqual("ERROR: Bad request", self.nodegroup.status_reason)

    @patch('magnum.drivers.common.driver.Driver.get_driver')
    def test_nodegroup_delete(self, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_get_driver.return_value = mock_driver
        self.handler.nodegroup_delete(self.context, self.cluster,
                                      self.nodegroup)
        mock_driver.delete_nodegroup.assert_called_once_with(self.context,
                                                             self.cluster,
                                                             self.nodegroup)
        self.assertEqual(fields.ClusterStatus.UPDATE_IN_PROGRESS,
                         self.cluster.status)
        self.assertEqual(fields.ClusterStatus.DELETE_IN_PROGRESS,
                         self.nodegroup.status)

    @patch('magnum.drivers.common.driver.Driver.get_driver')
    def test_nodegroup_delete_stack_not_found(self, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_get_driver.return_value = mock_driver
        nodegroup = mock.MagicMock()
        mock_driver.delete_nodegroup.side_effect = exc.HTTPNotFound()
        self.handler.nodegroup_delete(self.context, self.cluster, nodegroup)
        mock_driver.delete_nodegroup.assert_called_once_with(self.context,
                                                             self.cluster,
                                                             nodegroup)
        self.assertEqual(fields.ClusterStatus.UPDATE_IN_PROGRESS,
                         self.cluster.status)
        nodegroup.destroy.assert_called_once()

    @patch('magnum.drivers.common.driver.Driver.get_driver')
    def test_nodegroup_delete_stack_and_ng_not_found(self, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_get_driver.return_value = mock_driver
        nodegroup = mock.MagicMock()
        mock_driver.delete_nodegroup.side_effect = exc.HTTPNotFound()
        nodegroup.destroy.side_effect = exception.NodeGroupNotFound()
        self.handler.nodegroup_delete(self.context, self.cluster, nodegroup)
        mock_driver.delete_nodegroup.assert_called_once_with(self.context,
                                                             self.cluster,
                                                             nodegroup)
        self.assertEqual(fields.ClusterStatus.UPDATE_IN_PROGRESS,
                         self.cluster.status)
        nodegroup.destroy.assert_called_once()

    @patch('magnum.drivers.common.driver.Driver.get_driver')
    def test_nodegroup_delete_stack_operation_ongoing(self, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_get_driver.return_value = mock_driver
        mock_driver.delete_nodegroup.side_effect = exc.HTTPConflict()
        self.assertRaises(exception.NgOperationInProgress,
                          self.handler.nodegroup_delete,
                          self.context, self.cluster, self.nodegroup)
        mock_driver.delete_nodegroup.assert_called_once_with(self.context,
                                                             self.cluster,
                                                             self.nodegroup)
        self.assertEqual(fields.ClusterStatus.UPDATE_IN_PROGRESS,
                         self.cluster.status)
        self.assertEqual(fields.ClusterStatus.DELETE_IN_PROGRESS,
                         self.nodegroup.status)

    @patch('magnum.drivers.common.driver.Driver.get_driver')
    def test_nodegroup_delete_failed(self, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_get_driver.return_value = mock_driver
        side_effect = NotImplementedError("Delete failed")
        mock_driver.delete_nodegroup.side_effect = side_effect
        self.assertRaises(NotImplementedError,
                          self.handler.nodegroup_delete,
                          self.context, self.cluster, self.nodegroup)
        mock_driver.delete_nodegroup.assert_called_once_with(self.context,
                                                             self.cluster,
                                                             self.nodegroup)
        self.assertEqual(fields.ClusterStatus.UPDATE_FAILED,
                         self.cluster.status)
        self.assertEqual(fields.ClusterStatus.DELETE_FAILED,
                         self.nodegroup.status)
        self.assertEqual("Delete failed", self.nodegroup.status_reason)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging
from tempest.lib import exceptions

from magnum.tests.functional.api.v1.models import cluster_id_model
from magnum.tests.functional.api.v1.models import cluster_model
from magnum.tests.functional.common import client
from magnum.tests.functional.common import utils


class ClusterClient(client.MagnumClient):
    """Encapsulates REST calls and maps JSON to/from models"""

    LOG = logging.getLogger(__name__)

    @classmethod
    def clusters_uri(cls, filters=None):
        """Construct clusters uri with optional filters

        :param filters: Optional k:v dict that's converted to url query
        :returns: url string
        """

        url = "/clusters"
        if filters:
            url = cls.add_filters(url, filters)
        return url

    @classmethod
    def cluster_uri(cls, cluster_id):
        """Construct cluster uri

        :param cluster_id: cluster uuid or name
        :returns: url string
        """

        return "{0}/{1}".format(cls.clusters_uri(), cluster_id)

    def list_clusters(self, filters=None, **kwargs):
        """Makes GET /clusters request and returns ClusterCollection

        Abstracts REST call to return all clusters

        :param filters: Optional k:v dict that's converted to url query
        :returns: response object and ClusterCollection object
        """

        resp, body = self.get(self.clusters_uri(filters), **kwargs)
        return self.deserialize(resp, body, cluster_model.ClusterCollection)

    def get_cluster(self, cluster_id, **kwargs):
        """Makes GET /cluster request and returns ClusterEntity

        Abstracts REST call to return a single cluster based on uuid or name

        :param cluster_id: cluster uuid or name
        :returns: response object and ClusterCollection object
        """

        resp, body = self.get(self.cluster_uri(cluster_id))
        return self.deserialize(resp, body, cluster_model.ClusterEntity)

    def post_cluster(self, model, **kwargs):
        """Makes POST /cluster request and returns ClusterIdEntity

        Abstracts REST call to create new cluster

        :param model: ClusterEntity
        :returns: response object and ClusterIdEntity object
        """

        resp, body = self.post(
            self.clusters_uri(),
            body=model.to_json(), **kwargs)
        return self.deserialize(resp, body, cluster_id_model.ClusterIdEntity)

    def patch_cluster(self, cluster_id, clusterpatch_listmodel, **kwargs):
        """Makes PATCH /cluster request and returns ClusterIdEntity

        Abstracts REST call to update cluster attributes

        :param cluster_id: UUID of cluster
        :param clusterpatch_listmodel: ClusterPatchCollection
        :returns: response object and ClusterIdEntity object
        """

        resp, body = self.patch(
            self.cluster_uri(cluster_id),
            body=clusterpatch_listmodel.to_json(), **kwargs)
        return self.deserialize(resp, body, cluster_id_model.ClusterIdEntity)

    def delete_cluster(self, cluster_id, **kwargs):
        """Makes DELETE /cluster request and returns response object

        Abstracts REST call to delete cluster based on uuid or name

        :param cluster_id: UUID or name of cluster
        :returns: response object
        """

        return self.delete(self.cluster_uri(cluster_id), **kwargs)

    def wait_for_cluster_to_delete(self, cluster_id):
        utils.wait_for_condition(
            lambda: self.does_cluster_not_exist(cluster_id), 10, 600)

    def wait_for_created_cluster(self, cluster_id, delete_on_error=True):
        try:
            utils.wait_for_condition(
                lambda: self.does_cluster_exist(cluster_id), 10, 1800)
        except Exception:
            # In error state.  Clean up the cluster id if desired
            self.LOG.error('Cluster %s entered an exception state.',
                           cluster_id)
            if delete_on_error:
                self.LOG.error('We will attempt to delete clusters now.')
                self.delete_cluster(cluster_id)
                self.wait_for_cluster_to_delete(cluster_id)
            raise

    def wait_for_final_state(self, cluster_id):
        utils.wait_for_condition(
            lambda: self.is_cluster_in_final_state(cluster_id), 10, 1800)

    def is_cluster_in_final_state(self, cluster_id):
        try:
            resp, model = self.get_cluster(cluster_id)
            if model.status in ['CREATED', 'CREATE_COMPLETE',
                                'ERROR', 'CREATE_FAILED']:
                self.LOG.info('Cluster %s succeeded.', cluster_id)
                return True
            else:
                return False
        except exceptions.NotFound:
            self.LOG.warning('Cluster %s is not found.', cluster_id)
            return False

    def does_cluster_exist(self, cluster_id):
        try:
            resp, model = self.get_cluster(cluster_id)
            if model.status in ['CREATED', 'CREATE_COMPLETE']:
                self.LOG.info('Cluster %s is created.', cluster_id)
                return True
            elif model.status in ['ERROR', 'CREATE_FAILED']:
                self.LOG.error('Cluster %s is in fail state.',
                               cluster_id)
                raise exceptions.ServerFault(
                    "Got into an error condition: %s for %s",
                    (model.status, cluster_id))
            else:
                return False
        except exceptions.NotFound:
            self.LOG.warning('Cluster %s is not found.', cluster_id)
            return False

    def does_cluster_not_exist(self, cluster_id):
        try:
            self.get_cluster(cluster_id)
        except exceptions.NotFound:
            self.LOG.warning('Cluster %s is not found.', cluster_id)
            return True
        return False
